import json
from types import SimpleNamespace
from typing import Any, Dict
from unittest.mock import MagicMock, patch

import httpx
import pytest

from fbnconfig import configuration as cfg
from fbnconfig import identity, scheduler
from fbnconfig.identity import UserRef, UserResource, UserType


def response_hook(response: httpx.Response) -> None:
    if response.is_error:
        response.read()
        response.raise_for_status()


def mock_upload_response_aws(dest_name, dest_tag):
    return {
        "dockerLoginCommand":
            "docker login -u AWS -p TOK1 foo.lusid.com",
        "buildVersionedDockerImageCommand":
            f"docker build -t foo/{dest_name}:{dest_tag} . --platform=linux/amd64",
        "tagVersionedDockerImageCommand":
            f"docker tag foo/{dest_name}:{dest_tag} foo.lusid.com/foo/{dest_name}:{dest_tag}",
        "pushVersionedDockerImageCommand":
            f"docker push foo.lusid.com/foo/{dest_name}:{dest_tag}",
        "tagLatestDockerImageCommand":
            f"docker tag foo/{dest_name}:{dest_tag} foo.lusid.com/foo/{dest_name}:latest",
        "pushLatestDockerImageCommand":
            f"docker push foo.lusid.com/foo/{dest_name}:latest",
        "expiryTime": "2024-01-30T17:15:58.6730000+00:00",
    }


def mock_upload_response_az(dest_name, dest_tag):
    return {
      "dockerLoginCommand":
          "docker login -u fbn-rob-az-token -p TOK1 fbnciaz.azurecr.io",
      "buildVersionedDockerImageCommand":
          f"docker build -t foo/{dest_name}:{dest_tag} . --platform=linux/amd64",
      "tagVersionedDockerImageCommand":
          f"docker tag foo/{dest_name}:{dest_tag} fbnciaz.azurecr.io/foo/{dest_name}:{dest_tag}",
      "pushVersionedDockerImageCommand":
          f"docker push fbnciaz.azurecr.io/foo/{dest_name}:{dest_tag}",
      "tagLatestDockerImageCommand":
          f"docker tag foo/{dest_name}:{dest_tag} fbnciaz.azurecr.io/foo/{dest_name}:latest",
      "pushLatestDockerImageCommand":
          f"docker push fbnciaz.azurecr.io/foo/{dest_name}:latest",
       "expiryTime": "2024-01-30T17:15:58.6730000+00:00",
}


@pytest.mark.respx(base_url="https://foo.lusid.com")
class DescribeImageRef:
    base_url = "https://foo.lusid.com"
    client = httpx.Client(base_url=base_url, event_hooks={"response": [response_hook]})

    def test_attach_when_present(self, respx_mock):
        # given that the remote image exists with the required tag
        respx_mock.get("/scheduler2/api/images/repository/myimage").mock(
            return_value=httpx.Response(
                200,
                json={
                    "values": [
                        {
                            "name": "myimage",
                            "tags": [{"name": "0.1.0"}, {"name": "0.1.1"}, {"name": "latest"}],
                        }
                    ]
                },
            )
        )
        sut = scheduler.ImageRef(id="img1", dest_name="myimage", dest_tag="0.1.1")
        # when we call attach
        sut.attach(self.client)
        # then a get request was made and no exception raised

    def test_attach_when_missing_tag(self, respx_mock):
        # given the remote image exists but not with the required tag
        respx_mock.get("/scheduler2/api/images/repository/myimage").mock(
            return_value=httpx.Response(
                200,
                json={"values": [{"name": "myimage", "tags": [{"name": "0.1.0"}, {"name": "latest"}]}]},
            )
        )
        sut = scheduler.ImageRef(id="img1", dest_name="myimage", dest_tag="0.1.1")
        # when we call attach an exception is raised
        with pytest.raises(RuntimeError) as ex:
            sut.attach(self.client)
        assert "Image with name myimage and tag 0.1.1 not found" in str(ex.value)

    def test_attach_when_http_not_found(self, respx_mock):
        # given the server returns an error
        respx_mock.get("/scheduler2/api/images/repository/myimage").mock(
            return_value=httpx.Response(404, json={})
        )
        sut = scheduler.ImageRef(id="img1", dest_name="myimage", dest_tag="0.1.1")
        # when we call attach an exception is raised
        with pytest.raises(RuntimeError) as ex:
            sut.attach(self.client)
        assert "Image with name myimage and tag 0.1.1 not found" in str(ex.value)


@pytest.mark.respx(base_url="https://foo.lusid.com")
class DescribeImage:
    base_url = "https://foo.lusid.com"
    client = httpx.Client(base_url=base_url)

    @staticmethod
    def test_read():
        img = scheduler.ImageResource(
            id="image1", source_image="docker.io/alpine:5", dest_name="dest_alpine", dest_tag="v1"
        )
        img.read(httpx.Client(), SimpleNamespace())
        assert img.id == "image1"

    def test_delete(self):
        old_state = {
            "id": "image1",
            "source_image": "docker.io/alpine:5",
            "dest_name": "dest_alpine",
            "dest_tag": "v1",
        }
        # delete does nothing because we can't delete an image that's been uploaded
        sut = scheduler.ImageResource(
            id="image1", source_image="docker.io/alpine:5", dest_name="dest_alpine", dest_tag="v1"
        )
        sut.delete(self.client, old_state)

    @staticmethod
    def test_commands():
        sut = scheduler.ImageResource(
            id="image1", source_image="docker.io/alpine:5", dest_name="dest_alpine", dest_tag="v1"
        )
        user = "AWS"
        password = "TOK1"
        reghost = "foo.lusid.com"
        downstream_tag = "foo.lusid.com/foo/beany:v1"
        source_image = "docker.io/alpine:5"
        commands = (
            sut._pull_commands(source_image)
            + sut._tag_commands(downstream_tag)
            + sut._push_commands(downstream_tag, user, password, reghost)
        )
        assert commands == [
            ["docker", "pull", "--platform", "linux/amd64", "-q", "docker.io/alpine:5"],
            ["docker", "tag", "docker.io/alpine:5", "foo.lusid.com/foo/beany:v1"],
            ["docker", "login", "-u", "AWS", "--password", "TOK1", "foo.lusid.com"],
            ["docker", "push", "-q", "foo.lusid.com/foo/beany:v1"],
        ]

    @patch("subprocess.run")
    def test_create_no_matching_downstream_aws(self, mock_run, respx_mock):
        # given we have not pushed this image to an AWS lusid before
        def process_mock(args, **kwargs):
            if args[1] == "inspect":
                ret = MagicMock()
                ret.stdout = '["alpine@sha256:aaabbbccc"]'
                return ret
            return MagicMock()

        mock_run.side_effect = process_mock
        upload = mock_upload_response_aws("dest_alpine", "not_the_real_tag")
        respx_mock.post("/scheduler2/api/images").mock(side_effect=[httpx.Response(200, json=upload)])
        # when we create it
        sut = scheduler.ImageResource(
            id="image1", source_image="docker.io/alpine:5", dest_name="dest_alpine", dest_tag="v1"
        )
        new_state = sut.create(self.client)
        # then the new state is correct
        assert new_state == {
            "id": "image1",
            "source_image": "docker.io/alpine:5",
            "dest_name": "dest_alpine",
            "dest_tag": "v1",
        }
        # and the post method on scheduler gets the stock commands
        request = respx_mock.calls.last.request
        assert request.method == "POST"
        assert request.url == f"{self.base_url}/scheduler2/api/images"
        # and docker pull was called on the upstream
        assert len(mock_run.call_args_list) == 5
        pull_args = next(a[0][0] for a in mock_run.call_args_list if a[0][0][1] == "pull")
        assert pull_args == ["docker", "pull", "--platform", "linux/amd64", "-q", "docker.io/alpine:5"]
        # and the upstream was tagged with the downstream image
        tag_args = next(a[0][0] for a in mock_run.call_args_list if a[0][0][1] == "tag")
        assert tag_args == ["docker", "tag", "docker.io/alpine:5", "foo.lusid.com/foo/dest_alpine:v1"]
        # and docker push was called on the downstream image
        push_args = next(a[0][0] for a in mock_run.call_args_list if a[0][0][1] == "push")
        assert push_args == ["docker", "push", "-q", "foo.lusid.com/foo/dest_alpine:v1"]

    @patch("subprocess.run")
    def test_create_no_matching_downstream_az(self, mock_run, respx_mock):
        # given we have not pushed this image to an azure lusid before
        def process_mock(args, **kwargs):
            if args[1] == "inspect":
                ret = MagicMock()
                ret.stdout = '["alpine@sha256:aaabbbccc"]'
                return ret
            return MagicMock()

        mock_run.side_effect = process_mock
        # the az variant of the commands response
        upload = mock_upload_response_az("dest_alpine", "not_the_real_tag")
        respx_mock.post("/scheduler2/api/images").mock(side_effect=[httpx.Response(200, json=upload)])
        # when we create it
        sut = scheduler.ImageResource(
            id="image1", source_image="docker.io/alpine:5", dest_name="dest_alpine", dest_tag="v1"
        )
        new_state = sut.create(self.client)
        # then the new state is correct
        assert new_state == {
            "id": "image1",
            "source_image": "docker.io/alpine:5",
            "dest_name": "dest_alpine",
            "dest_tag": "v1",
        }
        # and the post method on scheduler gets the stock commands
        request = respx_mock.calls.last.request
        assert request.method == "POST"
        assert request.url == f"{self.base_url}/scheduler2/api/images"
        # and docker pull was called on the upstream
        assert len(mock_run.call_args_list) == 5
        pull_args = next(a[0][0] for a in mock_run.call_args_list if a[0][0][1] == "pull")
        assert pull_args == ["docker", "pull", "--platform", "linux/amd64", "-q", "docker.io/alpine:5"]
        # and the upstream was tagged with the downstream image which uses the container registry
        # host instead of the lusid domain
        tag_args = next(a[0][0] for a in mock_run.call_args_list if a[0][0][1] == "tag")
        assert tag_args == [
            "docker",
            "tag",
            "docker.io/alpine:5",
            "fbnciaz.azurecr.io/foo/dest_alpine:v1"
        ]
        # and docker push was called on the downstream image
        push_args = next(a[0][0] for a in mock_run.call_args_list if a[0][0][1] == "push")
        assert push_args == ["docker", "push", "-q", "fbnciaz.azurecr.io/foo/dest_alpine:v1"]

    @patch("subprocess.run")
    def test_create_no_matching_downstream_no_pull(self, mock_run, respx_mock):
        # given we have not pushed this image to lusid before
        def process_mock(args, **kwargs):
            if args[1] == "inspect":
                ret = MagicMock()
                ret.stdout = '["alpine@sha256:aaabbbccc"]'
                return ret
            return MagicMock()

        mock_run.side_effect = process_mock
        upload = mock_upload_response_aws("dest_alpine", "not_the_tag")
        respx_mock.post("/scheduler2/api/images").mock(side_effect=[httpx.Response(200, json=upload)])
        # given a new image with no pull so we upload the local image
        sut = scheduler.ImageResource(
            id="image1",
            source_image="alpine:5",
            dest_name="dest_alpine",
            dest_tag="v1",
            pull_upstream=False,
        )
        # when we create it
        new_state = sut.create(self.client)
        # then the new state is correct
        assert new_state == {
            "id": "image1",
            "source_image": "alpine:5",
            "dest_name": "dest_alpine",
            "dest_tag": "v1",
        }
        # and the post method on scheduler gets the stock commands
        request = respx_mock.calls.last.request
        assert request.method == "POST"
        assert request.url == f"{self.base_url}/scheduler2/api/images"
        # and docker pull was NOT called to fetch the usptream
        pull_args = [a[0][0] for a in mock_run.call_args_list if a[0][0][1] == "pull"]
        assert len(pull_args) == 0
        # and the existing local image was tagged with the downstream name
        tag_args = next(a[0][0] for a in mock_run.call_args_list if a[0][0][1] == "tag")
        assert tag_args == ["docker", "tag", "alpine:5", "foo.lusid.com/foo/dest_alpine:v1"]
        # and docker push was called on the upstream image
        assert len(mock_run.call_args_list) == 4
        push_args = next(a[0][0] for a in mock_run.call_args_list if a[0][0][1] == "push")
        assert push_args == ["docker", "push", "-q", "foo.lusid.com/foo/dest_alpine:v1"]

    @patch("subprocess.run")
    def test_create_image_already_exists_downstream(self, mock_run, respx_mock):
        # given we have pushed this image to lusid but it's not in the old_state
        # (maybe we deleted the state but the image can't be deleted)
        def process_mock(args, **kwargs):
            if args[1] == "inspect":
                ret = MagicMock()
                ret.stdout = (
                    '["alpine@sha256:aaabbbccc", "foo.lusid.com/foo/dest_alpine@sha256:123456abcdef"]'
                )
                return ret
            return MagicMock()

        mock_run.side_effect = process_mock
        # the api gives us the commands
        upload = mock_upload_response_aws("dest_alpine", "not_the_tag")
        respx_mock.post("/scheduler2/api/images").mock(side_effect=[httpx.Response(200, json=upload)])
        # the image exists in the downstream repo
        respx_mock.get("/scheduler2/api/images/repository/dest_alpine").mock(
            side_effect=[
                httpx.Response(
                    200,
                    json={
                        "values": [
                            {"digest": "sha256:7654321", "tags": [{"name": "not_mine"}]},
                            {"digest": "sha256:123456abcdef", "tags": [{"name": "v3"}, {"name": "v1"}]},
                        ]
                    },
                )
            ]
        )
        # given a new image
        sut = scheduler.ImageResource(
            id="image1", source_image="docker.io/alpine:5", dest_name="dest_alpine", dest_tag="v1"
        )
        # when we call create
        new_state = sut.create(self.client)
        # then the new state is correct
        assert new_state == {
            "id": "image1",
            "source_image": "docker.io/alpine:5",
            "dest_name": "dest_alpine",
            "dest_tag": "v1",
        }
        # the image with matching digest and tag in ecr is detected so
        #     we didn't upload the image with push
        #     we didn't call the post image endpoint
        assert len(mock_run.call_args_list) == 3

    @staticmethod
    def test_deps():
        # given an image
        sut = scheduler.ImageResource(
            id="image1", source_image="docker.io/alpine:5", dest_name="dest_alpine", dest_tag="v1"
        )
        # it doesn't have any dependencies
        assert sut.deps() == []

    def test_update_no_change(self):
        # given an existing job
        old_state = SimpleNamespace(
            id="image1", source_image="docker.io/alpine:5", dest_name="dest_alpine", dest_tag="v1"
        )
        # when we update with the same options
        sut = scheduler.ImageResource(
            id="image1", source_image="docker.io/alpine:5", dest_name="dest_alpine", dest_tag="v1"
        )
        new_state = sut.update(self.client, old_state)
        # then the state is None indicating no change
        assert new_state is None

    @patch("subprocess.run")
    def test_update_new_tag(self, mock_run, respx_mock):
        upload = mock_upload_response_aws("dest_alpine", "not_the_tag")
        respx_mock.post("/scheduler2/api/images").mock(side_effect=[httpx.Response(200, json=upload)])

        # given docker inspect will not return a downstream tag
        def process_mock(args, **kwargs):
            if args[1] == "inspect":
                ret = MagicMock()
                ret.stdout = '["alpine@sha256:aaabbbccc"]'
                return ret
            return MagicMock()

        mock_run.side_effect = process_mock
        # given an existing job using v0 of the image
        old_state = SimpleNamespace(
            id="image1", source_image="docker.io/alpine:5", dest_name="dest_alpine", dest_tag="v0"
        )
        # when we update to v1
        sut = scheduler.ImageResource(
            id="image1", source_image="docker.io/alpine:5", dest_name="dest_alpine", dest_tag="v1"
        )
        # then the new state is v1
        new_state = sut.update(self.client, old_state)
        assert new_state == {
            "id": "image1",
            "source_image": "docker.io/alpine:5",
            "dest_name": "dest_alpine",
            "dest_tag": "v1",
        }
        # and v1 gets pushed
        mock_run.assert_called_with(["docker", "push", "-q", "foo.lusid.com/foo/dest_alpine:v1"])

    def test_dump(self):
        # given an image resource
        sut = scheduler.ImageResource(
            id="img1", source_image="docker.io/alpine:3.16.7", dest_name="myimage", dest_tag="3.16.7"
        )
        # when we dump it
        dumped = sut.model_dump(
            by_alias=True, round_trip=True, exclude_none=True, context={"style": "dump"}
        )
        # then the dumped state is correct
        assert dumped == {
            "id": "img1",
            "sourceImage": "docker.io/alpine:3.16.7",
            "destName": "myimage",
            "destTag": "3.16.7",
        }

    def test_undump(self):
        # given a dumped image state
        dumped = {
            "id": "img1",
            "sourceImage": "docker.io/alpine:3.16.7",
            "destName": "myimage",
            "destTag": "3.16.7",
        }
        # when we undump it
        sut = scheduler.ImageResource.model_validate(
            dumped, context={"style": "dump", "$refs": {}, "id": "img1"}
        )
        # then the id has been extracted from the context
        assert sut.id == "img1"
        assert sut.source_image == "docker.io/alpine:3.16.7"
        assert sut.dest_name == "myimage"
        assert sut.dest_tag == "3.16.7"
        assert sut.pull_upstream is True  # default value

    def test_from_api(self, testdata_loader):
        # GIVEN an Image resource
        input_data = testdata_loader(
            "image_resource",
            "list_mock_data"
        )["values"][0]
        expected_result = testdata_loader(
            "image_resource",
            "list_expected_result"
        )["values"][0]["value"]
        # AND the required keys are populated
        input_data["destName"] = "ExampleName"
        input_data["destTag"] = "latest"
        # WHEN the from_api method is called for that resource
        result = scheduler.ImageResource.from_api(input_data)
        # THEN the response is as expected
        assert result == expected_result
        # AND model is validated successfully
        scheduler.ImageResource.model_validate(result)

    @pytest.mark.asyncio
    async def test_list(self, respx_mock, testdata_loader):
        # GIVEN an Image resource
        input_data = testdata_loader(
            "image_resource",
            "list_mock_data"
        )
        expected_result = testdata_loader(
            "image_resource",
            "list_expected_result"
        )
        # AND the resource exists in LUSID
        path_params = {
            "destName": "ExampleName"
        }
        url = scheduler.ImageResource.list_endpoint().format(**path_params)
        respx_mock.get(url).mock(
            return_value=httpx.Response(200, json=input_data)
        )
        # WHEN the list method is called for that resource
        async with httpx.AsyncClient(base_url=self.base_url) as client:
            result = await scheduler.ImageResource.list(client, path_params=path_params)
        # THEN the response is as expected
        assert result == expected_result

    @pytest.mark.asyncio
    async def test_fetch(self, respx_mock, testdata_loader):
        # GIVEN an Image resource
        input_data = testdata_loader(
            "image_resource",
            "list_mock_data"
        )["values"][0]
        expected_result = testdata_loader(
            "image_resource",
            "list_expected_result"
        )["values"][0]["value"]
        # AND the resource exists in LUSID
        keys = {
            "destName": input_data["name"],
            "destTag": input_data["tags"][0]["name"],
        }
        url = scheduler.ImageResource.fetch_endpoint().format(**keys)
        respx_mock.get(url).mock(
            return_value=httpx.Response(200, json=input_data)
        )
        # WHEN the fetch method is called for that resource
        async with httpx.AsyncClient(base_url=self.base_url) as client:
            result = await scheduler.ImageResource.fetch(client, keys=keys)
        # THEN the response is as expected
        assert result == expected_result
        # AND the correct request was made to LUSID
        request = respx_mock.calls.last.request
        assert request.method == "GET"
        assert request.url.path == url
        assert request.url.query == b""


@pytest.mark.respx(base_url="https://foo.lusid.com")
class DescribeJobRef:
    base_url = "https://foo.lusid.com"
    client = httpx.Client(base_url=base_url)

    def test_attach_when_remote_exists(self, respx_mock):
        # given that a job is returned from the list call
        respx_mock.get("/scheduler2/api/jobs").mock(
            side_effect=[
                httpx.Response(200, json={"values": [{"name": "job name", "argumentDefinitions": {}}]})
            ]
        )
        # when we create a JobRef
        sut = scheduler.JobRef(id="10", scope="scope_a", code="code_a")
        sut.attach(self.client)
        # then no exception

    def test_attach_when_remote_exists_with_argument_definition(self, respx_mock):
        # given that a job is returned from the list call
        respx_mock.get("/scheduler2/api/jobs").mock(
            side_effect=[
                httpx.Response(
                    200,
                    json={
                        "values": [
                            {
                                "name": "job name",
                                "argumentDefinitions": {
                                    "someargument": {
                                        "dataType": "String",
                                        "required": True,
                                        "description": "my argument",
                                        "order": 1,
                                        "constraints": "None",
                                        "passedAs": "CommandLine",
                                        "defaultValue": "somedefaultvalue",
                                    }
                                },
                            }
                        ]
                    },
                )
            ]
        )

        # when we create a JobRef
        sut = scheduler.JobRef(id="10", scope="scope_a", code="code_a")
        sut.attach(self.client)
        # then no exception

    def test_attach_when_no_remote(self, respx_mock):
        # given that a job is returned from the list call
        respx_mock.get("/scheduler2/api/jobs").mock(
            side_effect=[httpx.Response(200, json={"values": []})]
        )
        # when we create a JobRef
        with pytest.raises(RuntimeError):
            sut = scheduler.JobRef(id="10", scope="scope_a", code="code_a")
            sut.attach(self.client)
            # then an exception is thrown


@pytest.mark.respx(base_url="https://foo.lusid.com")
class DescribeJob:
    def setup_class(self):
        scheduler.JobResource.ScanTimeoutParameters.tries = 3
        scheduler.JobResource.ScanTimeoutParameters.wait_time = 0

    base_url = "https://foo.lusid.com"
    jobs_url = "/scheduler2/api/jobs"

    client = httpx.Client(base_url=base_url, event_hooks={"response": [response_hook]})

    @pytest.fixture
    def existing_image(self):
        return scheduler.ImageResource(
            id="image1", source_image="docker.io/alpine:5", dest_name="dest_alpine", dest_tag="v1"
        )

    @pytest.fixture
    def image_scan_stub(self):
        return self.image_scan_value("dest_alpine", "v1", "COMPLETE")

    @pytest.fixture
    def config_item_ref(self):
        set = cfg.SetRef(id="set", scope="cfgscope", code="cfgcode", type=cfg.SetType.PERSONAL)
        item = cfg.ItemRef(id="item", set=set, key="itemkey")
        # fake the item being attached
        item.ref = "config://123"
        return item

    def test_create_with_config_arg(self, respx_mock, existing_image, config_item_ref, image_scan_stub):
        respx_mock.get(f"/scheduler2/api/images/repository/{existing_image.dest_name}").mock(
            side_effect=httpx.Response(200, json={"values": [image_scan_stub]})
        )
        # given an existing image
        # and a config item that's already got a ref
        # when we create a job using the config item as a default arg value
        sut = scheduler.JobResource(
            id="job1",
            scope="scope1",
            code="code1",
            image=existing_image,
            name="job name",
            description="a job",
            argument_definitions={
                "myargument1": scheduler.EnvironmentArg(
                    data_type="String",
                    required=True,
                    description="a description",
                    order=1,
                    default_value=config_item_ref,
                ),
                "myargument2": scheduler.CommandlineArg(
                    data_type="Int",
                    required=False,
                    description="other description",
                    order=2,
                    default_value="3",
                ),
            },
        )
        respx_mock.post("/scheduler2/api/jobs").mock(
            side_effect=[httpx.Response(200, json={"name": "job name"})]
        )
        new_state = sut.create(self.client)
        # the state is created
        assert new_state == {"id": "job1", "scope": "scope1", "code": "code1"}
        # and the post is made to create the job passing the config
        # item ref in the argument
        request = respx_mock.calls.last.request
        assert request.method == "POST"
        assert request.url == f"{self.base_url}/scheduler2/api/jobs"
        assert json.loads(request.content) == {
            "jobId": {"scope": "scope1", "code": "code1"},
            "imageName": "dest_alpine",
            "imageTag": "v1",
            "name": "job name",
            "description": "a job",
            "argumentDefinitions": {
                "myargument1": {
                    "dataType": "String",
                    "required": True,
                    "description": "a description",
                    "order": 1,
                    "passedAs": "EnvironmentVariable",
                    "defaultValue": "config://123",
                },
                "myargument2": {
                    "dataType": "Int",
                    "required": False,
                    "description": "other description",
                    "order": 2,
                    "passedAs": "CommandLine",
                    "defaultValue": "3",
                },
            },
        }

    def test_create(self, respx_mock, existing_image, image_scan_stub):
        respx_mock.get(f"/scheduler2/api/images/repository/{existing_image.dest_name}").mock(
            side_effect=httpx.Response(200, json={"values": [image_scan_stub]})
        )

        # given an existing image
        # when we create a job
        sut = scheduler.JobResource(
            id="job1",
            scope="scope1",
            code="code1",
            image=existing_image,
            name="job name",
            description="a job",
            argument_definitions={
                "myargument1": scheduler.EnvironmentArg(
                    data_type="String",
                    required=True,
                    description="a description",
                    order=1,
                    default_value=None,
                ),
                "myargument2": scheduler.CommandlineArg(
                    data_type="Int",
                    required=False,
                    description="other description",
                    order=2,
                    default_value="3",
                ),
            },
        )

        respx_mock.post("/scheduler2/api/jobs").mock(
            side_effect=[httpx.Response(200, json={"name": "job name"})]
        )

        new_state = sut.create(self.client)
        # the state is created
        assert new_state == {"id": "job1", "scope": "scope1", "code": "code1"}

        # and the post is made to create the job
        request = respx_mock.calls.last.request
        assert request.method == "POST"
        assert request.url == f"{self.base_url}/scheduler2/api/jobs"
        assert json.loads(request.content) == {
            "jobId": {"scope": "scope1", "code": "code1"},
            "imageName": "dest_alpine",
            "imageTag": "v1",
            "name": "job name",
            "description": "a job",
            "argumentDefinitions": {
                "myargument1": {
                    "dataType": "String",
                    "required": True,
                    "description": "a description",
                    "order": 1,
                    "passedAs": "EnvironmentVariable",
                },
                "myargument2": {
                    "dataType": "Int",
                    "required": False,
                    "description": "other description",
                    "order": 2,
                    "passedAs": "CommandLine",
                    "defaultValue": "3",
                },
            },
        }

    def test_update_no_change(self, respx_mock, existing_image):
        respx_mock.get("/scheduler2/api/jobs").mock(
            side_effect=[
                httpx.Response(
                    200,
                    json={
                        "values": [
                            {
                                "name": "job name",
                                "jobId": {},
                                "dockerImage": "dest_alpine:v1",
                                "description": "a job",
                                "minCpu": "1",
                                "maxCpu": "2",
                                "dateCreated": "2023-01-02",
                                "requiredResources": {
                                    "lusidApis": [],
                                    "lusidFileSystem": [],
                                    "externalCalls": [],
                                },
                                "commandLineArgumentSeparator": " ",
                                "author": "bob",
                                "argumentDefinitions": {
                                    "arg1": {
                                        "dataType": "String",
                                        "required": True,
                                        "description": "some",
                                        "order": 1,
                                        "constraints": "None",
                                        "passedAs": "CommandLine",
                                    }
                                },
                            }
                        ]
                    },
                )
            ]
        )
        old_state = SimpleNamespace(id="job1", scope="scope1", code="code1")
        sut = scheduler.JobResource(
            id="job1",
            scope="scope1",
            code="code1",
            image=existing_image,
            name="job name",
            description="a job",
            argument_definitions={
                "arg1": scheduler.CommandlineArg(
                    data_type="String", required=True, description="some", order=1
                )
            },
        )
        new_state = sut.update(self.client, old_state)
        assert new_state is None

    def test_update_leaves_default_fields(self, respx_mock, existing_image):
        # given a job with an author (a field which is defaulted by the api)
        respx_mock.get("/scheduler2/api/jobs").mock(
            side_effect=[
                httpx.Response(
                    200,
                    json={
                        "values": [
                            {
                                "name": "job name",
                                "jobId": {},
                                "author": "somebody",
                                "dockerImage": "dest_alpine:v1",
                                "description": "a job",
                                "requiredResources": {
                                    "lusidApis": [],
                                    "lusidFileSystem": [],
                                    "externalCalls": [],
                                },
                                "commandLineArgumentSeparator": " ",
                                "argumentDefinitions": {},
                            }
                        ]
                    },
                )
            ]
        )
        # when we update and leave the author blank
        old_state = SimpleNamespace(id="job1", scope="scope1", code="code1")
        sut = scheduler.JobResource(
            id="job1",
            scope="scope1",
            code="code1",
            image=existing_image,
            name="job name",
            description="a job",
        )
        new_state = sut.update(self.client, old_state)
        # then no request is made so we preserve the default
        assert new_state is None

    def test_update_modify_default_field(self, respx_mock, existing_image, image_scan_stub):
        # given a job with an author (a field which is defaulted by the api)
        respx_mock.get(f"/scheduler2/api/images/repository/{existing_image.dest_name}").mock(
            side_effect=httpx.Response(200, json={"values": [image_scan_stub]})
        )

        respx_mock.get("/scheduler2/api/jobs").mock(
            side_effect=[
                httpx.Response(
                    200,
                    json={
                        "values": [
                            {
                                "name": "job name",
                                "jobId": {},
                                "author": "somebody",
                                "dockerImage": "dest_alpine:v1",
                                "description": "a job",
                                "requiredResources": {
                                    "lusidApis": [],
                                    "lusidFileSystem": [],
                                    "externalCalls": [],
                                },
                                "commandLineArgumentSeparator": " ",
                                "argumentDefinitions": {},
                            }
                        ]
                    },
                )
            ]
        )
        respx_mock.put("/scheduler2/api/jobs/scope1/code1").mock(
            side_effect=[httpx.Response(200, json={"name": "different job name"})]
        )
        # when we provide a value for author
        old_state = SimpleNamespace(id="job1", scope="scope1", code="code1")
        sut = scheduler.JobResource(
            id="job1",
            scope="scope1",
            code="code1",
            image=existing_image,
            name="job name",
            description="a job",
            author="someone else",
        )
        new_state = sut.update(self.client, old_state)
        # then the new state the same as the old one
        assert new_state == {"id": "job1", "scope": "scope1", "code": "code1"}
        # and a request is made to change the author
        request = respx_mock.calls.last.request
        assert request.method == "PUT"
        assert json.loads(request.content)["author"] == "someone else"

    def test_update_default_field_same_value(self, respx_mock, existing_image):
        # given a job with an author (a field which is defaulted by the api)
        respx_mock.get("/scheduler2/api/jobs").mock(
            side_effect=[
                httpx.Response(
                    200,
                    json={
                        "values": [
                            {
                                "name": "job name",
                                "jobId": {},
                                "author": "somebody",
                                "dockerImage": "dest_alpine:v1",
                                "description": "a job",
                                "requiredResources": {
                                    "lusidApis": [],
                                    "lusidFileSystem": [],
                                    "externalCalls": [],
                                },
                                "commandLineArgumentSeparator": " ",
                                "argumentDefinitions": {},
                            }
                        ]
                    },
                )
            ]
        )
        # when the desired author is specified but is the same as the remote
        old_state = SimpleNamespace(id="job1", scope="scope1", code="code1")
        sut = scheduler.JobResource(
            id="job1",
            scope="scope1",
            code="code1",
            image=existing_image,
            name="job name",
            description="a job",
            author="somebody",
        )
        new_state = sut.update(self.client, old_state)
        # then the new state is None because no modification is required
        assert new_state is None
        # and no put request is made to change

    def test_update_change_name(self, respx_mock, existing_image, image_scan_stub):
        # given an existing job with name 'a job'
        respx_mock.get(f"/scheduler2/api/images/repository/{existing_image.dest_name}").mock(
            side_effect=httpx.Response(200, json={"values": [image_scan_stub]})
        )

        respx_mock.get("/scheduler2/api/jobs").mock(
            side_effect=[
                httpx.Response(
                    200,
                    json={
                        "values": [
                            {
                                "name": "job name",
                                "jobId": {},
                                "author": "bob",
                                "dockerImage": "dest_alpine:v1",
                                "description": "a job",
                                "requiredResources": {
                                    "lusidApis": [],
                                    "lusidFileSystem": [],
                                    "externalCalls": [],
                                },
                                "commandLineArgumentSeparator": " ",
                                "argumentDefinitions": {},
                            }
                        ]
                    },
                )
            ]
        )
        respx_mock.put("/scheduler2/api/jobs/scope1/code1").mock(
            side_effect=[httpx.Response(200, json={"name": "different job name"})]
        )
        # when we update to 'different job name'
        old_state = SimpleNamespace(id="job1", scope="scope1", code="code1")
        sut = scheduler.JobResource(
            id="job1",
            scope="scope1",
            code="code1",
            image=existing_image,
            name="different job name",
            description="a job",
        )
        new_state = sut.update(self.client, old_state)
        # then the new state is created
        assert new_state == {"id": "job1", "scope": "scope1", "code": "code1"}
        # and the post is made to create the job
        request = respx_mock.calls.last.request
        assert request.method == "PUT"
        assert request.url == f"{self.base_url}/scheduler2/api/jobs/scope1/code1"
        assert json.loads(request.content) == {
            "argumentDefinitions": {},
            "description": "a job",
            "name": "different job name",
            "imageName": "dest_alpine",
            "imageTag": "v1",
        }

    def test_update_changed_argument_names(self, respx_mock, existing_image, image_scan_stub):
        # given an existing job arg1 only
        respx_mock.get(f"/scheduler2/api/images/repository/{existing_image.dest_name}").mock(
            side_effect=httpx.Response(200, json={"values": [image_scan_stub]})
        )
        respx_mock.get("/scheduler2/api/jobs").mock(
            side_effect=[
                httpx.Response(
                    200,
                    json={
                        "values": [
                            {
                                "name": "job name",
                                "jobId": {},
                                "author": "bob",
                                "dockerImage": "dest_alpine:v1",
                                "description": "a job",
                                "requiredResources": {
                                    "lusidApis": [],
                                    "lusidFileSystem": [],
                                    "externalCalls": [],
                                },
                                "commandLineArgumentSeparator": " ",
                                "argumentDefinitions": {
                                    "arg1": {
                                        "dataType": "String",
                                        "required": True,
                                        "description": "some",
                                        "order": 1,
                                        "constraints": "None",
                                        "passedAs": "CommandLine",
                                    }
                                },
                            }
                        ]
                    },
                )
            ]
        )
        respx_mock.put("/scheduler2/api/jobs/scope1/code1").mock(
            side_effect=[httpx.Response(200, json={"name": "different job name"})]
        )
        old_state = SimpleNamespace(id="job1", scope="scope1", code="code1")
        # when we update with a desired state where arg1 is removed and arg2 is added
        sut = scheduler.JobResource(
            id="job1",
            scope="scope1",
            code="code1",
            image=existing_image,
            name="job name",
            description="a job",
            argument_definitions={
                "arg2": scheduler.EnvironmentArg(
                    data_type="Configuration", required=True, description="some", order=1
                )
            },
        )
        # then it is detected as a change and arg2 is sent
        new_state = sut.update(self.client, old_state)
        assert new_state == {"id": "job1", "scope": "scope1", "code": "code1"}
        request = respx_mock.calls.last.request
        assert request.method == "PUT"
        assert request.url == f"{self.base_url}/scheduler2/api/jobs/scope1/code1"
        assert json.loads(request.content) == {
            "argumentDefinitions": {
                "arg2": {
                    "dataType": "Configuration",
                    "required": True,
                    "description": "some",
                    "order": 1,
                    "passedAs": "EnvironmentVariable",
                }
            },
            "description": "a job",
            "name": "job name",
            "imageName": "dest_alpine",
            "imageTag": "v1",
        }

    def test_update_when_removing_default_value(self, respx_mock, existing_image, image_scan_stub):
        respx_mock.get(f"/scheduler2/api/images/repository/{existing_image.dest_name}").mock(
            side_effect=httpx.Response(200, json={"values": [image_scan_stub]})
        )
        # given an existing job with arg1 that has a default value
        respx_mock.get("/scheduler2/api/jobs").mock(
            side_effect=[
                httpx.Response(
                    200,
                    json={
                        "values": [
                            {
                                "name": "job name",
                                "jobId": {},
                                "author": "bob",
                                "dockerImage": "dest_alpine:v1",
                                "description": "a job",
                                "requiredResources": {
                                    "lusidApis": [],
                                    "lusidFileSystem": [],
                                    "externalCalls": [],
                                },
                                "commandLineArgumentSeparator": " ",
                                "argumentDefinitions": {
                                    "arg1": {
                                        "dataType": "String",
                                        "required": True,
                                        "description": "some",
                                        "order": 1,
                                        "constraints": "None",
                                        "passedAs": "EnvironmentVariable",
                                        "defaultValue": "somevalue",
                                    }
                                },
                            }
                        ]
                    },
                )
            ]
        )

        respx_mock.put("/scheduler2/api/jobs/scope1/code1").mock(
            side_effect=[httpx.Response(200, json={"name": "different job name"})]
        )
        old_state = SimpleNamespace(id="job1", scope="scope1", code="code1")
        # when we update with a desired state where arg1 default value is removed
        sut = scheduler.JobResource(
            id="job1",
            scope="scope1",
            code="code1",
            image=existing_image,
            name="job name",
            description="a job",
            argument_definitions={
                "arg1": scheduler.EnvironmentArg(
                    data_type="String", required=True, description="some", order=1
                )
            },
        )
        # then it is detected as a change and arg1 is sent
        new_state = sut.update(self.client, old_state)
        assert new_state == {"id": "job1", "scope": "scope1", "code": "code1"}
        request = respx_mock.calls.last.request
        assert request.method == "PUT"
        assert request.url == f"{self.base_url}/scheduler2/api/jobs/scope1/code1"
        assert json.loads(request.content) == {
            "argumentDefinitions": {
                "arg1": {
                    "dataType": "String",
                    "required": True,
                    "description": "some",
                    "order": 1,
                    "passedAs": "EnvironmentVariable",
                }
            },
            "description": "a job",
            "name": "job name",
            "imageName": "dest_alpine",
            "imageTag": "v1",
        }

    def test_update_merges_args(self, respx_mock, existing_image, image_scan_stub):
        # given an existing job
        respx_mock.get(f"/scheduler2/api/images/repository/{existing_image.dest_name}").mock(
            side_effect=httpx.Response(200, json={"values": [image_scan_stub]})
        )

        respx_mock.get("/scheduler2/api/jobs").mock(
            side_effect=[
                httpx.Response(
                    200,
                    json={
                        "values": [
                            {
                                "name": "job name",
                                "jobId": {},
                                "author": "bob",
                                "dockerImage": "dest_alpine:v1",
                                "description": "a job",
                                "requiredResources": {
                                    "lusidApis": [],
                                    "lusidFileSystem": [],
                                    "externalCalls": [],
                                },
                                "commandLineArgumentSeparator": " ",
                                "argumentDefinitions": {
                                    "arg1": {
                                        "dataType": "String",
                                        "required": True,
                                        "description": "some",
                                        "order": 1,
                                        "constraints": "None",
                                        "passedAs": "CommandLine",
                                    }
                                },
                            }
                        ]
                    },
                )
            ]
        )
        respx_mock.put("/scheduler2/api/jobs/scope1/code1").mock(
            side_effect=[httpx.Response(200, json={"name": "different job name"})]
        )
        old_state = SimpleNamespace(id="job1", scope="scope1", code="code1")
        # when we update with a field inside the argument changed (required=False)
        sut = scheduler.JobResource(
            id="job1",
            scope="scope1",
            code="code1",
            image=existing_image,
            name="job name",
            description="a job",
            argument_definitions={
                "arg1": scheduler.CommandlineArg(
                    data_type="String", required=False, description="some", order=1
                )
            },
        )
        # then it is detected as a change and the modified arg is sent
        new_state = sut.update(self.client, old_state)
        assert new_state == {"id": "job1", "scope": "scope1", "code": "code1"}
        request = respx_mock.calls.last.request
        assert request.method == "PUT"
        assert request.url == f"{self.base_url}/scheduler2/api/jobs/scope1/code1"
        assert json.loads(request.content) == {
            "argumentDefinitions": {
                "arg1": {
                    "dataType": "String",
                    "required": False,
                    "description": "some",
                    "order": 1,
                    "passedAs": "CommandLine",
                }
            },
            "description": "a job",
            "name": "job name",
            "imageName": "dest_alpine",
            "imageTag": "v1",
        }

    def test_deps(self, existing_image):
        sut = scheduler.JobResource(
            id="job1",
            scope="scope1",
            code="code1",
            image=existing_image,
            name="different job name",
            description="a job",
        )
        assert sut.deps() == [existing_image]

    def test_deps_with_config_item(self, existing_image, config_item_ref):
        sut = scheduler.JobResource(
            id="job1",
            scope="scope1",
            code="code1",
            image=existing_image,
            name="different job name",
            description="a job",
            argument_definitions={
                "myargument1": scheduler.EnvironmentArg(
                    data_type="String",
                    required=True,
                    description="a description",
                    order=1,
                    default_value=config_item_ref,
                )
            },
        )
        assert sut.deps() == [existing_image, config_item_ref]

    @staticmethod
    def image_scan_value(name: str, tag: str, scan_status: str) -> Dict[Any, Any]:
        return {
            "name": name,
            "pushTime": "2024-11-22T15:11:39.9570000+00:00",
            "digest": "sha256:0e5733c4ffc5317a24dcd10adbe5dbe0499f641a3b13a74a118ac5059385f605",
            "size": 586206179,
            "tags": [
                {
                    "name": tag,
                    "pullTime": "0001-01-01T00:00:00.0000000+00:00",
                    "pushTime": "0001-01-01T00:00:00.0000000+00:00",
                    "signed": False,
                    "immutable": False,
                },
                {
                    "name": "00000",
                    "pullTime": "0001-01-01T00:00:00.0000000+00:00",
                    "pushTime": "0001-01-01T00:00:00.0000000+00:00",
                    "signed": False,
                    "immutable": False,
                },
            ],
            "scanStatus": scan_status,
            "scanSummary": {},
        }

    def test_vulnerability_scan_multiple_fail(self, respx_mock, existing_image):
        # given: there are mutliple images in the repository.
        # An old one which # was scanned ok,
        # the one this job is referencing which failed the scan
        dest_name = existing_image.dest_name
        dest_tag = existing_image.dest_tag
        respx_mock.get("/scheduler2/api/images/repository/dest_alpine").mock(
            side_effect=[
                httpx.Response(
                    200,
                    json={
                        "values": [
                            self.image_scan_value(dest_name, dest_tag, "FAILED"),
                            self.image_scan_value(dest_name, "other_tag", "COMPLETE"),
                        ]
                    },
                )
            ]
        )
        # when we check the scan
        sut = scheduler.JobResource(
            id="job1",
            scope="scope1",
            code="code1",
            image=existing_image,
            name="job name",
            description="a job",
            argument_definitions={},
        )
        # then it fails because the image/tag combination this job uses has failed the scan
        with pytest.raises(RuntimeError, match="Image vulnerability scan failed."):
            sut._wait_for_vulnerability_scan(self.client)

    def test_vulnerability_scan_multiple_success(self, respx_mock, existing_image):
        # given: there are mutliple images in the repository. An old one which
        # failed the scan, and the one this job is referencing which completes
        dest_name = existing_image.dest_name
        respx_mock.get("/scheduler2/api/images/repository/dest_alpine").mock(
            side_effect=[
                httpx.Response(
                    200,
                    json={
                        "values": [
                            self.image_scan_value(dest_name, existing_image.dest_tag, "COMPLETE"),
                            self.image_scan_value(dest_name, "other_tag", "FAILED"),
                        ]
                    },
                )
            ]
        )
        # when we check the scan
        sut = scheduler.JobResource(
            id="job1",
            scope="scope1",
            code="code1",
            image=existing_image,
            name="job name",
            description="a job",
            argument_definitions={},
        )
        # then it completes because the referenced image is ok
        sut._wait_for_vulnerability_scan(self.client)

    def test_vulnerability_scan_waits_for_complete(self, respx_mock, existing_image):
        # given the scan is in progress at the time the job is created
        # and then completes with a successful scan
        respx_mock.get("/scheduler2/api/images/repository/dest_alpine").mock(
            side_effect=[
                httpx.Response(
                    200,
                    json={
                        "values": [
                            self.image_scan_value(existing_image.dest_name, existing_image.dest_tag, "")
                        ]
                    },
                ),
                httpx.Response(
                    200,
                    json={
                        "values": [
                            self.image_scan_value(
                                existing_image.dest_name, existing_image.dest_tag, "IN_PROGRESS"
                            )
                        ]
                    },
                ),
                httpx.Response(
                    200,
                    json={
                        "values": [
                            self.image_scan_value(
                                existing_image.dest_name, existing_image.dest_tag, "COMPLETE"
                            )
                        ]
                    },
                ),
            ]
        )
        # when we check the scan
        sut = scheduler.JobResource(
            id="job1",
            scope="scope1",
            code="code1",
            image=existing_image,
            name="job name",
            description="a job",
            argument_definitions={},
        )
        # then it is successful because the final state of the scan is COMPLETE
        sut._wait_for_vulnerability_scan(self.client)

    @pytest.mark.parametrize(
        "terminal_state",
        ["FAILED", "UNSUPPORTED_IMAGE", "SCAN_ELIGIBILITY_EXPIRED", "FINDINGS_UNAVAILABLE"],
    )
    def test_wait_for_vulnerability_scan_failure(self, respx_mock, existing_image, terminal_state):
        # given the scan is in progress when the job is created and results
        # in any of the failure states
        respx_mock.get("/scheduler2/api/images/repository/dest_alpine").mock(
            side_effect=[
                httpx.Response(
                    200,
                    json={
                        "values": [
                            self.image_scan_value(existing_image.dest_name, existing_image.dest_tag, "")
                        ]
                    },
                ),
                httpx.Response(
                    200,
                    json={
                        "values": [
                            self.image_scan_value(
                                existing_image.dest_name, existing_image.dest_tag, "IN_PROGRES"
                            )
                        ]
                    },
                ),
                httpx.Response(
                    200,
                    json={
                        "values": [
                            self.image_scan_value(
                                existing_image.dest_name, existing_image.dest_tag, terminal_state
                            )
                        ]
                    },
                ),
            ]
        )
        # when we check the scan
        sut = scheduler.JobResource(
            id="job1",
            scope="scope1",
            code="code1",
            image=existing_image,
            name="job name",
            description="a job",
            argument_definitions={},
        )
        # then an exception is raised which causes the job update to fail
        with pytest.raises(RuntimeError, match="Image vulnerability scan failed."):
            sut._wait_for_vulnerability_scan(self.client)

    def test_wait_for_vulnerability_scan_timeout(self, respx_mock, existing_image):
        # given the scan is in progress and stays in progress for 4 requests
        # where our scan tries is 3
        respx_mock.get("/scheduler2/api/images/repository/dest_alpine").mock(
            side_effect=[
                httpx.Response(
                    200,
                    json={
                        "values": [
                            self.image_scan_value(
                                existing_image.dest_name, existing_image.dest_tag, "IN_PROGRESS"
                            )
                        ]
                    },
                )
            ]
            * 4
        )
        # when we wait for the scan
        sut = scheduler.JobResource(
            id="job1",
            scope="scope1",
            code="code1",
            image=existing_image,
            name="job name",
            description="a job",
            argument_definitions={},
        )
        # then a timeout exception is raised because the scan did not complete in time
        with pytest.raises(RuntimeError, match="Image vulnerability scan timed out."):
            sut._wait_for_vulnerability_scan(self.client)

    def test_dump_with_config_arg(self, existing_image, config_item_ref):
        # given a resource
        sut = scheduler.JobResource(
            id="job1",
            scope="scope1",
            code="code1",
            image=existing_image,
            name="job name",
            description="a job",
            argument_definitions={
                "myargument1": scheduler.EnvironmentArg(
                    data_type="String",
                    required=True,
                    description="a description",
                    order=1,
                    default_value=config_item_ref,
                ),
                "myargument2": scheduler.CommandlineArg(
                    data_type="Int",
                    required=False,
                    description="other description",
                    order=2,
                    default_value="3",
                ),
            },
        )
        # when we dump it
        dumped = sut.model_dump(
            by_alias=True, exclude_none=True, round_trip=True, context={"style": "dump"}
        )
        # then:
        assert dumped == {
            "scope": "scope1",  # it has scope/code instead of jobId
            "code": "code1",
            "image": {"$ref": existing_image.id},  # the image is a $ref
            "name": "job name",
            "description": "a job",
            "argumentDefinitions": {
                "myargument1": {
                    "dataType": "String",
                    "required": True,
                    "description": "a description",
                    "order": 1,
                    "passedAs": "EnvironmentVariable",
                    "defaultValue": {"$ref": config_item_ref.id},  # config item is $ref
                },
                "myargument2": {
                    "dataType": "Int",
                    "required": False,
                    "description": "other description",
                    "order": 2,
                    "passedAs": "CommandLine",
                    "defaultValue": "3",
                },
            },
        }

    def test_undump(self, existing_image, config_item_ref):
        # given a dumped resource
        dumped = {
            "scope": "scope1",
            "code": "code1",
            "image": {"$ref": existing_image.id},
            "name": "job name",
            "description": "a job",
            "argumentDefinitions": {
                "myargument1": {
                    "dataType": "String",
                    "required": True,
                    "description": "a description",
                    "order": 1,
                    "passedAs": "EnvironmentVariable",
                    "defaultValue": {"$ref": config_item_ref.id},
                },
                "myargument2": {
                    "dataType": "Int",
                    "required": False,
                    "description": "other description",
                    "order": 2,
                    "passedAs": "CommandLine",
                    "defaultValue": "3",
                },
            },
        }
        # when we undump it
        sut = scheduler.JobResource.model_validate(
            dumped,
            context={
                "style": "dump",
                "id": "job1",
                "$refs": {existing_image.id: existing_image, config_item_ref.id: config_item_ref},
            },
        )
        # then scope/code are read
        assert sut.scope == dumped["scope"]
        assert sut.code == dumped["code"]
        # the image is wired up to the resource
        assert sut.image == existing_image
        # the id field has been extracted from the context
        assert sut.id == "job1"
        # the config $ref has been wired up to the config item provided
        assert sut.argument_definitions["myargument1"].default_value == config_item_ref
        assert sut.name == dumped["name"]
        assert sut.description == dumped["description"]

    def test_parse_api_format_list_response(self, existing_image, config_item_ref):
        # given a job in the same format as list jobs returns (jobId.scope and jobId.code)
        # and "dockerImage" as a single item
        # but with refs for the related objects
        resp = {
            "jobId": {"scope": "scope1", "code": "code1"},
            "dockerImage": {"$ref": existing_image.id},
            "name": "job name",
            "description": "a job",
            "argumentDefinitions": {
                "myargument1": {
                    "dataType": "String",
                    "required": True,
                    "description": "a description",
                    "order": 1,
                    "passedAs": "EnvironmentVariable",
                    "defaultValue": {"$ref": config_item_ref.id},
                },
                "myargument2": {
                    "dataType": "Int",
                    "required": False,
                    "description": "other description",
                    "order": 2,
                    "passedAs": "CommandLine",
                    "defaultValue": "3",
                },
            },
        }
        # when we parse it
        sut = scheduler.JobResource.model_validate(
            resp,
            context={
                "$refs": {existing_image.id: existing_image, config_item_ref.id: config_item_ref},
                "id": "job1",
            },
        )
        # then the scope and code are from the jobId
        assert sut.scope == "scope1"
        assert sut.code == "code1"
        # and the first argument is an env var
        first_arg = sut.argument_definitions["myargument1"]
        assert isinstance(first_arg, scheduler.EnvironmentArg)
        assert isinstance(first_arg.default_value, cfg.ItemRef)
        # with the config ref resolved
        assert first_arg.default_value.id == config_item_ref.id
        # and the second is a commandline
        assert isinstance(sut.argument_definitions["myargument2"], scheduler.CommandlineArg)

    def test_from_api(self, testdata_loader):
        data = testdata_loader(
            "job_resource",
            "list_mock_data"
        )["values"][0]
        expected_data = testdata_loader(
            "job_resource",
            "list_expected_result"
        )["values"][0]["value"]
        result = scheduler.JobResource.from_api(data)
        assert result == expected_data
        image_resource = testdata_loader(
            "image_resource",
            "list_expected_result"
        )["values"][0]["value"]
        image_model = scheduler.ImageResource(**image_resource)
        job_model = scheduler.JobResource.model_validate(result, context={
            "$refs": {image_model.id: image_model}
        })
        assert job_model.image == image_model

    @pytest.mark.asyncio
    async def test_list(self, respx_mock, testdata_loader):
        data = testdata_loader(
            "job_resource",
            "list_mock_data"
        )
        expected_data = testdata_loader(
            "job_resource",
            "list_expected_result"
        )
        respx_mock.get(scheduler.JobResource.list_endpoint()).mock(
            return_value=httpx.Response(200, json=data)
        )
        async with httpx.AsyncClient(base_url=self.base_url) as client:
            result = await scheduler.JobResource.list(client, query_params={})
        assert result == expected_data
        # Verify the correct request was made
        assert respx_mock.calls.last.request.method == "GET"
        assert "/scheduler2/api/jobs" in str(respx_mock.calls.last.request.url)


@pytest.mark.respx(base_url="https://foo.lusid.com")
class DescribeSchedule:
    base_url = "https://foo.lusid.com"
    schedules_url = "/scheduler2/api/schedules"

    client = httpx.Client(base_url=base_url, event_hooks={"response": [response_hook]})

    @pytest.fixture
    def existing_job(self):
        return scheduler.JobResource(
            id="job1",
            scope="scope1",
            code="code1",
            image=scheduler.ImageResource(
                id="image1", source_image="docker.io/alpine:5", dest_name="dest_alpine", dest_tag="v1"
            ),
            name="job name",
            description="a job",
            argument_definitions={
                "myarg1": scheduler.CommandlineArg(data_type="String", description="test", order=1)
            },
        )

    @pytest.fixture
    def config_item_ref(self):
        set = cfg.SetRef(id="set", scope="cfgscope", code="cfgcode", type=cfg.SetType.PERSONAL)
        item = cfg.ItemRef(id="item", set=set, key="itemkey")
        # fake the item being attached
        item.ref = "config://123"
        return item

    def test_read_given_schedule_exists(self, respx_mock, existing_job: scheduler.JobResource):
        schedule = scheduler.ScheduleResource(
            id="myschedule",
            name="name",
            scope="schedulescope",
            code="schedulecode",
            expression="somecron",
            timezone="Europe/London",
            job=existing_job,
            description="description",
        )
        old_state = SimpleNamespace(scope="scope1", code="code1")
        respx_mock.get(f"{self.schedules_url}/{old_state.scope}/{old_state.code}").mock(
            side_effect=[
                httpx.Response(
                    200, json={"somefield": "myschedule", "scheduleIdentifier": "somethingelse"}
                )
            ]
        )
        # when read is called
        # the remote is populated but scheduleIdentifier is not included
        assert schedule.read(client=self.client, old_state=old_state) == {"somefield": "myschedule"}

    def test_create_with_config_arg(self, respx_mock, existing_job, config_item_ref):
        # given the schedule does not exists in the remote
        respx_mock.post(self.schedules_url).mock(
            side_effect=[httpx.Response(201, json={"some": "response"})]
        )
        # and the desired schedule references a config item
        schedule = scheduler.ScheduleResource(
            id="myschedule",
            name="testname",
            scope="scopes",
            code="codes",
            expression="somecron",
            job=existing_job,
            description="some description",
            arguments={"arg1": config_item_ref},
        )
        # when we create it
        schedule.create(self.client)
        # then the config item is turned into it's ref
        request = respx_mock.calls.last.request
        assert json.loads(request.content) == {
            "scheduleId": {"scope": "scopes", "code": "codes"},
            "jobId": {"scope": existing_job.scope, "code": existing_job.code},
            "trigger": {"timeTrigger": {"expression": "somecron"}},
            "name": "testname",
            "description": "some description",
            "arguments": {"arg1": "config://123"},
        }

    def test_update_with_config_arg(self, respx_mock, existing_job, config_item_ref):
        # GIVEN a remote schedule which does not use a config
        respx_mock.get(f"{self.schedules_url}/scope1/code1").mock(
            side_effect=[
                httpx.Response(
                    200,
                    json={
                        "scheduleIdentifier": "somethingelse",
                        "enabled": False,
                        "author": "default",
                        "owner": "default",
                        "useAsAuth": "default",
                        "notifications": "default",
                        "arguments": {"arg1": "astring"},
                    },
                )
            ]
        )
        respx_mock.put(f"{self.schedules_url}/scope1/code1").mock(
            side_effect=[httpx.Response(200, json={})]
        )
        # and the desired schedule references a config item
        schedule = scheduler.ScheduleResource(
            id="myschedule",
            name="testname",
            scope="scope1",
            code="code1",
            expression="somecron",
            timezone="Europe/London",
            job=existing_job,
            description="some description",
            arguments={"arg1": config_item_ref},
        )
        # when we update it
        old_state = SimpleNamespace(scope="scope1", code="code1")
        new_state = schedule.update(self.client, old_state)
        # then the config item is turned into it's ref in the put request
        request = respx_mock.calls.last.request
        assert json.loads(request.content) == {
            "jobId": {"scope": existing_job.scope, "code": existing_job.code},
            "trigger": {"timeTrigger": {"expression": "somecron", "timeZone": "Europe/London"}},
            "name": "testname",
            "description": "some description",
            "arguments": {"arg1": "config://123"},
        }
        # and a new state is returned
        assert new_state is not None

    @pytest.mark.parametrize(
        "user_resource",
        [
            UserResource(
                id="id",
                first_name="name",
                last_name="last",
                email_address="email",
                login="login",
                type=UserType.SERVICE,
            ),
            UserRef(id="id", login="login"),
        ],
    )
    def test_create_given_schedule_has_defaults(
        self, respx_mock, existing_job: scheduler.JobResource, user_resource
    ):
        schedule_scope = "testscope"
        schedule_code = "testcode"
        user = user_resource
        user.user_id = "someuserid"
        schedule = scheduler.ScheduleResource(
            id="myschedule",
            name="testname",
            scope=schedule_scope,
            code=schedule_code,
            expression="somecron",
            timezone="Europe/London",
            job=existing_job,
            description="some description",
            author="custom author",
            use_as_auth=user,
        )

        expected_post = {
            "scheduleId": {"scope": schedule_scope, "code": schedule_code},
            "jobId": {"scope": existing_job.scope, "code": existing_job.code},
            "trigger": {"timeTrigger": {"expression": "somecron", "timeZone": "Europe/London"}},
            "name": "testname",
            "author": "custom author",
            "description": "some description",
            "useAsAuth": "someuserid",
        }

        respx_mock.post(self.schedules_url).mock(
            side_effect=[httpx.Response(201, json={"some": "response"})]
        )

        response = schedule.create(client=self.client)

        assert (
            response
            == scheduler.ScheduleState(
                id="myschedule", scope=schedule_scope, code=schedule_code, argKeys=[]
            ).model_dump()
        )

        request = respx_mock.calls.last.request
        assert json.loads(request.content) == expected_post

    def test_update_given_id_is_changed_then_fails(self, existing_job: scheduler.JobResource):
        # GIVEN some schedule state
        schedule_scope = "testscope"
        schedule_code = "testcode"
        schedule = scheduler.ScheduleResource(
            id="myschedule",
            name="testname",
            scope=schedule_scope,
            code=schedule_code,
            expression="somecron",
            timezone="Europe/London",
            job=existing_job,
            description="some description",
        )

        def assert_result(scope: str, code: str):
            # WHEN the old state ID is different from current state
            old_state = SimpleNamespace(scope=scope, code=code)

            # THEN Update fails
            with pytest.raises(RuntimeError) as ex:
                schedule.update(client=self.client, old_state=old_state)
            assert f"{schedule_scope}/{schedule_code}" in str(ex.value)
            assert f"{old_state.scope}/{old_state.code}" in str(ex.value)

        # just the code is different
        assert_result(scope=schedule_scope, code="someothercode")
        # just the scope is different
        assert_result(scope="someotherscope", code=schedule_code)
        # scope and code are both different
        assert_result(scope="someotherscope", code="someothercode")

    def test_update_given_no_defaults_provided_then_handles_correctly(
        self, respx_mock, existing_job: scheduler.JobResource
    ):
        schedule_scope = "testscope"
        schedule_code = "testcode"

        old_state = SimpleNamespace(scope=schedule_scope, code=schedule_code)

        # GIVEN a remote state containing the defaults
        respx_mock.get(f"{self.schedules_url}/{old_state.scope}/{old_state.code}").mock(
            side_effect=[
                httpx.Response(
                    200,
                    json={
                        "scheduleIdentifier": "somethingelse",
                        "enabled": False,
                        "author": "default",
                        "owner": "default",
                        "useAsAuth": "default",
                        "notifications": "default",
                    },
                )
            ]
        )
        json = {
            "description": "some description",
            "jobId": {"code": existing_job.code, "scope": existing_job.scope},
            "name": "testname",
            "trigger": {"timeTrigger": {"expression": "somecron", "timeZone": "Europe/London"}},
        }
        respx_mock.put(f"{self.schedules_url}/{schedule_scope}/{schedule_code}", json=json).mock(
            side_effect=[
                httpx.Response(
                    200,
                    json={
                        "somefield": "myschedule",
                        "scheduleIdentifier": "somethingelse",
                        "enabled": False,
                        "author": "default",
                        "owner": "default",
                        "useAsAuth": "default",
                    },
                )
            ]
        )

        # AND a resource without the defaults and some changed fields
        schedule = scheduler.ScheduleResource(
            id="myschedule",
            name="testname",
            scope=schedule_scope,
            code=schedule_code,
            expression="somecron",
            timezone="Europe/London",
            job=existing_job,
            description="some description",
        )
        # WHEN Calling update
        # THEN Updated as expected
        state = schedule.update(client=self.client, old_state=old_state)
        assert state is not None
        assert state["id"] == "myschedule"

    def test_update_remove_argument(self, respx_mock, existing_job):
        # GIVEN A remote with two arguments
        respx_mock.get(f"{self.schedules_url}/scope1/code1").mock(
            side_effect=[
                httpx.Response(
                    200,
                    json={
                        "scheduleIdentifier": {"scope": "scop1", "code": "code1"},
                        "jobId": {"scope": existing_job.scope, "code": existing_job.code},
                        "trigger": {
                            "timeTrigger": {"expression": "somecron", "timeZone": "Europe/London"}
                        },
                        "name": "testname",
                        "author": "custom author",
                        "description": "some description",
                        "arguments": {"arg1": "argvalue1", "arg2": "argvalue2"},
                    },
                )
            ]
        )
        # and desired is the same but has removed one arg
        schedule = scheduler.ScheduleResource(
            id="myschedule",
            name="testname",
            scope="scope1",
            code="code1",
            expression="somecron",
            timezone="Europe/London",
            job=existing_job,
            description="some description",
            arguments={"arg1": "argvalue1"},
        )
        # when it gets updated
        respx_mock.put(f"{self.schedules_url}/scope1/code1").mock(
            side_effect=[httpx.Response(200, json={})]
        )
        old_state = SimpleNamespace(scope="scope1", code="code1", argKeys=["arg1", "arg2"])
        new_state = schedule.update(self.client, old_state)
        # then the change is detected and a state is returned
        assert new_state is not None
        assert new_state["argKeys"] == ["arg1"]

    @pytest.mark.parametrize(
        "user_resource",
        [
            UserResource(
                id="id",
                first_name="name",
                last_name="last",
                email_address="email",
                login="login",
                type=UserType.SERVICE,
            ),
            UserRef(id="id", login="login"),
        ],
    )
    def test_update_given_all_defaults_provided_then_handles_correctly(
        self, respx_mock, existing_job, user_resource
    ):
        schedule_scope = "testscope"
        schedule_code = "testcode"
        auth_user = user_resource
        auth_user.user_id = "notdefaultuser"

        old_state = SimpleNamespace(scope=schedule_scope, code=schedule_code)
        # GIVEN A remote state with all defaults set
        respx_mock.get(f"{self.schedules_url}/{old_state.scope}/{old_state.code}").mock(
            side_effect=[
                httpx.Response(
                    200,
                    json={
                        "somefield": "myschedule",
                        "scheduleIdentifier": "somethingelse",
                        "enabled": True,
                        "author": "notdefault",
                        "owner": "notdefault",
                        "useAsAuth": "notdefault",
                        "notifications": "some",
                        "arguments": {"myarg1": "myvalue1"},
                    },
                )
            ]
        )

        expected_update = {
            "description": "some description",
            "jobId": {"code": existing_job.code, "scope": existing_job.scope},
            "name": "testname",
            "trigger": {"timeTrigger": {"expression": "somecron", "timeZone": "Europe/London"}},
            "arguments": {"myarg2": "myvalue2"},
            "author": "notdefaultauthor",
            "owner": "notdefaultowner",
            "useAsAuth": "notdefaultuser",
        }

        respx_mock.put(
            f"{self.schedules_url}/{schedule_scope}/{schedule_code}", json=expected_update
        ).mock(
            side_effect=[
                httpx.Response(
                    200,
                    json={
                        "somefield": "myschedule",
                        "scheduleIdentifier": "somethingelse",
                        "enabled": False,
                        "author": "default",
                        "owner": "default",
                        "useAsAuth": "default",
                    },
                )
            ]
        )

        # AND A resource with all the defaults set and they are different
        schedule = scheduler.ScheduleResource(
            id="myschedule",
            name="testname",
            scope=schedule_scope,
            code=schedule_code,
            expression="somecron",
            timezone="Europe/London",
            job=existing_job,
            description="some description",
            arguments={"myarg2": "myvalue2"},
            author="notdefaultauthor",
            owner="notdefaultowner",
            use_as_auth=auth_user,
        )

        # WHEN Calling update
        state = schedule.update(client=self.client, old_state=old_state)
        assert state is not None
        assert state["id"] == "myschedule"

    def test_update_given_no_change_then_no_action(self, respx_mock):
        schedule_scope = "testscope"
        schedule_code = "testcode"
        user = UserRef(id="id", login="login")
        user.user_id = "notdefault"

        job = scheduler.JobResource(
            id="job1",
            scope="scope1",
            code="code1",
            image=scheduler.ImageResource(
                id="image1", source_image="docker.io/alpine:5", dest_name="dest_alpine", dest_tag="v1"
            ),
            name="job name",
            description="a job",
            argument_definitions={
                "myarg1": scheduler.CommandlineArg(data_type="String", description="test", order=1),
                "jobdefaultarg": scheduler.CommandlineArg(
                    data_type="String",
                    description="test",
                    order=1,
                    required=False,
                    default_value="somedefaultvalue",
                ),
            },
        )

        old_state = SimpleNamespace(
            scope=schedule_scope, code=schedule_code, argKeys=["myarg1", "jobdefaultarg"]
        )

        # GIVEN A remote that matches the resource
        respx_mock.get(f"{self.schedules_url}/{old_state.scope}/{old_state.code}").mock(
            side_effect=[
                httpx.Response(
                    200,
                    json={
                        "scheduleIdentifier": "somethingelse",
                        "description": "some description",
                        "jobId": {"code": job.code, "scope": job.scope},
                        "name": "testname",
                        "enabled": True,
                        "author": "notdefault",
                        "owner": "notdefault",
                        "useAsAuth": "notdefault",
                        "notifications": "some",
                        "trigger": {
                            "timeTrigger": {"expression": "somecron", "timeZone": "Europe/London"}
                        },
                        "arguments": {"myarg1": "myvalue1", "jobdefaultarg": "somedefaultvalue"},
                    },
                )
            ]
        )
        schedule = scheduler.ScheduleResource(
            id="myschedule",
            name="testname",
            scope=schedule_scope,
            code=schedule_code,
            expression="somecron",
            timezone="Europe/London",
            job=job,
            description="some description",
            arguments={"myarg1": "myvalue1", "jobdefaultarg": "somedefaultvalue"},
            author="notdefault",
            owner="notdefault",
            use_as_auth=user,
        )

        # WHEN Calling Update
        result = schedule.update(client=self.client, old_state=old_state)
        # Then None is returned
        assert result is None

    def test_delete_given_schedule_exists(self, respx_mock, existing_job):
        schedule_scope = "testscope"
        schedule_code = "testcode"
        schedule = scheduler.ScheduleResource(
            id="myschedule",
            name="testname",
            scope=schedule_scope,
            code=schedule_code,
            expression="somecron",
            timezone="Europe/London",
            job=existing_job,
            description="something",
        )

        old_state = SimpleNamespace(scope=schedule_scope, code=schedule_code)

        # GIVEN Delete returns 204
        respx_mock.delete(f"{self.schedules_url}/{schedule_scope}/{schedule_code}").mock(
            side_effect=[httpx.Response(204)]
        )

        # THEN Calling delete does not raise an exception
        schedule.delete(client=self.client, old_state=old_state)

    def test_delete_given_schedule_does_not_exist(self, respx_mock, existing_job):
        schedule_scope = "testscope"
        schedule_code = "testcode"
        schedule = scheduler.ScheduleResource(
            id="myschedule",
            name="testname",
            scope=schedule_scope,
            code=schedule_code,
            expression="somecron",
            timezone="Europe/London",
            job=existing_job,
            description="something",
        )

        old_state = SimpleNamespace(scope=schedule_scope, code=schedule_code)

        # GIVEN Delete returns 404 with name ValidationError
        # AND title Schedule could not be found
        respx_mock.delete(f"{self.schedules_url}/{schedule_scope}/{schedule_code}").mock(
            side_effect=[
                httpx.Response(
                    404, json={"name": "ValidationError", "title": "Schedule could not be found"}
                )
            ]
        )

        # THEN Calling delete does not raise an exception
        schedule.delete(client=self.client, old_state=old_state)

    def test_delete_given_failure_then_raises(self, respx_mock, existing_job):
        schedule_scope = "testscope"
        schedule_code = "testcode"
        schedule = scheduler.ScheduleResource(
            id="myschedule",
            name="testname",
            scope=schedule_scope,
            code=schedule_code,
            expression="somecron",
            timezone="Europe/London",
            job=existing_job,
            description="something",
        )

        old_state = SimpleNamespace(scope=schedule_scope, code=schedule_code)

        # GIVEN Delete returns 404 with some name or error title other than not found
        respx_mock.delete(f"{self.schedules_url}/{schedule_scope}/{schedule_code}").mock(
            side_effect=[
                httpx.Response(404, json={"name": "notavalidationerror", "title": "Something else"})
            ]
        )

        # THEN Calling delete does raise an exception
        with pytest.raises(httpx.HTTPStatusError) as error:
            schedule.delete(client=self.client, old_state=old_state)

        assert error.value.response.status_code == 404
        assert error.value.response.json()["name"] == "notavalidationerror"

    @staticmethod
    @pytest.mark.parametrize(
        "user",
        [
            UserResource(
                id="id",
                first_name="name",
                last_name="last",
                email_address="email",
                login="login",
                type=UserType.SERVICE,
            ),
            UserRef(id="id", login="login"),
        ],
    )
    def test_deps_with_config_and_user(config_item_ref, existing_job, user):
        # given a schedule which references a config item in its arguments
        schedule = scheduler.ScheduleResource(
            id="myschedule",
            name="testname",
            scope="scopes",
            code="codes",
            expression="somecron",
            timezone="Europe/London",
            job=existing_job,
            description="some description",
            arguments={"arg1": config_item_ref},
            use_as_auth=user,
        )
        # when we get the dependencies
        deps = schedule.deps()
        # then the deps includes the config item and the job
        assert deps == [existing_job, config_item_ref, user]

    def test_dump(self, existing_job):
        # given a simple schedule
        sut = scheduler.ScheduleResource(
            id="schedule1",
            name="Test Schedule",
            scope="scope1",
            code="code1",
            expression="0 9 * * *",
            timezone="Europe/London",
            job=existing_job,
            description="A test schedule",
            enabled=True,
            author="test-author",
            owner="test-owner",
        )
        # when we dump it
        dumped = sut.model_dump(
            by_alias=True,
            round_trip=True,  # excludes computeds
            exclude_none=True,
            context={"style": "dump"},
        )
        # then the dumped state is correct
        assert dumped == {
            "name": "Test Schedule",
            "description": "A test schedule",
            "scope": "scope1",  # scope/code are used instead of jobId
            "code": "code1",
            "expression": "0 9 * * *",
            "timezone": "Europe/London",
            "job": {"$ref": existing_job.id},
            "enabled": True,
            "author": "test-author",
            "owner": "test-owner",
        }

    def test_dump_with_config(self, existing_job, config_item_ref):
        # given an existing schedule with config arguments
        sut = scheduler.ScheduleResource(
            id="schedule1",
            name="Test Schedule",
            scope="scope1",
            code="code1",
            expression="0 9 * * *",
            timezone="Europe/London",
            job=existing_job,
            description="A test schedule",
            arguments={"myarg": config_item_ref, "stringarg": "test-value"},
        )
        # when we dump it
        dumped = sut.model_dump(
            by_alias=True,
            round_trip=True,  # excludes computeds
            exclude_none=True,
            context={"style": "dump"},
        )
        # then:
        assert dumped == {
            "scope": "scope1",
            "code": "code1",
            "name": "Test Schedule",
            "description": "A test schedule",
            "expression": "0 9 * * *",
            "timezone": "Europe/London",
            "job": {"$ref": existing_job.id},  # the job is a $ref
            "arguments": {
                "myarg": {"$ref": config_item_ref.id},  # the config item is a $ref
                "stringarg": "test-value",
            },
        }

    def test_undump_with_config_arg(self, existing_job, config_item_ref):
        # given a dumped schedule state
        dumped = {
            "scope": "scope1",
            "code": "code1",
            "name": "Test Schedule",
            "expression": "0 9 * * *",
            "timezone": "Europe/London",
            "description": "A test schedule",
            "job": {"$ref": existing_job.id},
            "enabled": True,
            "author": "test-author",
            "owner": "test-owner",
            "arguments": {"myarg": {"$ref": config_item_ref.id}, "stringarg": "test-value"},
        }
        # when we undump it
        sut = scheduler.ScheduleResource.model_validate(
            dumped,
            context={
                "style": "undump",
                "$refs": {existing_job.id: existing_job, config_item_ref.id: config_item_ref},
                "id": "schedule1",
            },
        )
        # then the id has been extracted from the context
        assert sut.id == "schedule1"
        assert sut.name == "Test Schedule"
        assert sut.description == "A test schedule"
        # and the job ref has been wired up
        assert sut.job == existing_job
        assert sut.enabled is True
        assert sut.author == "test-author"
        assert sut.owner == "test-owner"
        # and the config item ref has been wired up
        assert sut.arguments
        assert sut.arguments["myarg"] == config_item_ref
        assert sut.arguments["stringarg"] == "test-value"

    def test_from_api(self, testdata_loader):
        # GIVEN a Schedule resource
        input_data = testdata_loader(
            "schedule_resource",
            "list_mock_data"
        )["values"][0]
        expected_result = testdata_loader(
            "schedule_resource",
            "list_expected_result"
        )["values"][0]["value"]
        # WHEN the from_api method is called for that resource
        result = scheduler.ScheduleResource.from_api(input_data)
        # THEN the response is as expected
        assert result == expected_result
        # AND model is validated successfully
        image_resource = testdata_loader(
            "image_resource",
            "list_expected_result"
        )["values"][0]["value"]
        image_model = scheduler.ImageResource(**image_resource)
        job_resource = testdata_loader(
            "job_resource",
            "list_expected_result"
        )["values"][0]["value"]
        job_model = scheduler.JobResource.model_validate(job_resource, context={
            "$refs": {
                image_model.id: image_model
            }
        })
        user_resource = testdata_loader(
            "user_resource",
            "list_expected_result"
        )["values"][0]["value"]
        user_model = identity.UserResource(**user_resource)

        schedule_model = scheduler.ScheduleResource.model_validate(result, context={
            "$refs": {
                job_model.id: job_model,
                user_model.id: user_model
            }
        })
        assert schedule_model.job == job_model
        assert schedule_model.use_as_auth == user_model

    @pytest.mark.asyncio
    async def test_list(self, respx_mock, testdata_loader):
        # GIVEN a Schedule resource
        input_data = testdata_loader(
            "schedule_resource",
            "list_mock_data"
        )
        expected_result = testdata_loader(
            "schedule_resource",
            "list_expected_result"
        )
        # AND the resource exists in LUSID
        path_params = {}
        url = scheduler.ScheduleResource.list_endpoint().format(**path_params)
        respx_mock.get(url).mock(
            return_value=httpx.Response(200, json=input_data)
        )
        # WHEN the list method is called for that resource
        async with httpx.AsyncClient(base_url=self.base_url) as client:
            result = await scheduler.ScheduleResource.list(client, path_params=path_params)
        # THEN the response is as expected
        assert result == expected_result

    @pytest.mark.asyncio
    async def test_fetch(self, respx_mock, testdata_loader):
        # GIVEN a Schedule resource
        input_data = testdata_loader(
            "schedule_resource",
            "list_mock_data"
        )["values"][0]
        expected_result = testdata_loader(
            "schedule_resource",
            "list_expected_result"
        )["values"][0]["value"]
        # AND the resource exists in LUSID
        keys = {
            "scope": input_data["scheduleIdentifier"]["scope"],
            "code": input_data["scheduleIdentifier"]["code"],
        }
        url = scheduler.ScheduleResource.fetch_endpoint().format(**keys)
        respx_mock.get(url).mock(
            return_value=httpx.Response(200, json=input_data)
        )
        # WHEN the fetch method is called for that resource
        async with httpx.AsyncClient(base_url=self.base_url) as client:
            result = await scheduler.ScheduleResource.fetch(client, keys=keys)
        # THEN the response is as expected
        assert result == expected_result
        # AND the correct request was made to LUSID
        request = respx_mock.calls.last.request
        assert request.method == "GET"
        assert request.url.path == url
        assert request.url.query == b""


@pytest.mark.respx(base_url="https://foo.lusid.com")
class DescribeScheduleRef:
    base_url = "https://foo.lusid.com"
    client = httpx.Client(base_url=base_url, event_hooks={"response": [response_hook]})

    def test_attach_when_present(self, respx_mock):
        # given that the remote schedule exists
        respx_mock.get("/scheduler2/api/schedules/testscope/testcode").mock(
            return_value=httpx.Response(200, json={"name": "Test Schedule"})
        )
        sut = scheduler.ScheduleRef(id="sched1", scope="testscope", code="testcode")
        # when we call attach
        sut.attach(self.client)
        # then a get request was made and no exception raised

    def test_attach_when_missing(self, respx_mock):
        # given the remote schedule does not exist
        respx_mock.get("/scheduler2/api/schedules/testscope/testcode").mock(
            return_value=httpx.Response(404, json={})
        )
        sut = scheduler.ScheduleRef(id="sched1", scope="testscope", code="testcode")
        # when we call attach an exception is raised
        with pytest.raises(RuntimeError) as ex:
            sut.attach(self.client)
        assert "Schedule testscope/testcode not found" in str(ex.value)

    def test_attach_when_http_error(self, respx_mock):
        # given the server returns an error
        respx_mock.get("/scheduler2/api/schedules/testscope/testcode").mock(
            return_value=httpx.Response(500, json={})
        )
        sut = scheduler.ScheduleRef(id="sched1", scope="testscope", code="testcode")
        # when we call attach an exception is raised
        with pytest.raises(httpx.HTTPStatusError):
            sut.attach(self.client)



import contextlib
import importlib.util
import os
import secrets
import string
import sys


@contextlib.contextmanager
def add_to_path(p):
    import sys

    old_path = sys.path
    sys.path = sys.path[:]
    sys.path.insert(0, p)
    try:
        yield
    finally:
        sys.path = old_path


def gensym(prefix, length=32):
    """
    generates a fairly unique symbol, used to make a module name,
    used as a helper function for load_module

    :return: generated symbol
    """
    prefix = prefix or "gensym_"
    alphabet = string.ascii_uppercase + string.ascii_lowercase + string.digits
    symbol = "".join([secrets.choice(alphabet) for i in range(length)])
    return prefix + symbol


def load_module(source, module_name=None):
    """
    reads file source and loads it as a module

    :param source: file to load
    :param module_name: name of module to register in sys.modules
    :return: loaded module
    """
    module_name = gensym(module_name)
    # see: https://stackoverflow.com/a/41904558 for why we add the source
    # folder to the path here
    with add_to_path(os.path.dirname(source)):
        spec = importlib.util.spec_from_file_location(module_name, source)
        if spec is None or spec.loader is None:
            raise RuntimeError("Failed to get spec for module")
        module = importlib.util.module_from_spec(spec)
        sys.modules[module_name] = module
        spec.loader.exec_module(module)
        return module



from __future__ import annotations

from enum import StrEnum
from types import SimpleNamespace
from typing import Annotated, Any, List

import httpx
from httpx import Client as httpxClient
from pydantic import (
    BaseModel,
    BeforeValidator,
    Field,
    PlainSerializer,
    WithJsonSchema,
    computed_field,
    model_validator,
)

from fbnconfig import datatype

from .resource_abc import CamelAlias, FetchMixin, ListMixin, Ref, Resource, register_resource

ResourceId = datatype.ResourceId
DataTypeKey = datatype.DataTypeKey

GROUP = "Data Model"


class FieldCategory(StrEnum):
    SERIES_IDENTIFIER = "SeriesIdentifier"
    VALUE = "Value"
    METADATA = "Metadata"


class ApplicableEntityType(StrEnum):
    INSTRUMENT = "Instrument"
    LEGAL_ENTITY = "LegalEntity"
    PERSON = "Person"
    INVESTOR_RECORD = "InvestorRecord"
    INVESTMENT_ACCOUNT = "InvestmentAccount"
    PORTFOLIO = "Portfolio"
    TRANSACTION = "Transaction"
    HOLDING = "Holding"
    SETTLEMENT_INSTRUCTION = "SettlementInstruction"


class FieldSchema(BaseModel, CamelAlias):
    """Define a field in the relational dataset schema"""
    field_name: str
    display_name: str | None = None
    description: str | None = None
    data_type_id: DataTypeKey
    required: bool = True
    category:  FieldCategory


@register_resource(group=GROUP)
class RelationalDatasetDefinitionRef(BaseModel, Ref):
    """Reference an existing relational dataset definition

    Example
    -------
    >>> from fbnconfig import relational_dataset_definition
    >>> relational_dataset_definition.RelationalDatasetDefinitionRef(
    ...     id="sample-dataset-ref",
    ...     scope="default",
    ...     code="sample-dataset"
    ... )

    Attributes
    ----------
    id : str
        Resource identifier.
    scope : str
        Scope of the dataset definition.
    code : str
        Code of the dataset definition.
    """

    id: str = Field(exclude=True)
    scope: str
    code: str

    def attach(self, client):
        try:
            client.get(f"/api/api/relationaldatasetdefinitions/{self.scope}/{self.code}")
        except httpx.HTTPStatusError as ex:
            if ex.response.status_code == 404:
                raise RuntimeError(
                    f"Relational dataset definition {self.scope}/{self.code} does not exist"
                )
            else:
                raise ex


@register_resource(group=GROUP, keys=["scope", "code"])
class RelationalDatasetDefinitionResource(BaseModel, Resource, ListMixin, FetchMixin, CamelAlias):
    """Define a relational dataset definition

    Example
    -------
    >>> from fbnconfig import relational_dataset_definition, datatype
    >>> relational_dataset_definition.RelationalDatasetDefinitionResource(
    ...     id="sample-dataset",
    ...     scope="default",
    ...     code="sample-dataset",
    ...     display_name="Sample Financial Dataset",
    ...     description="A sample relational dataset definition",
    ...     applicable_entity_types=["Instrument", "Portfolio"],
    ...     field_schema=[
    ...         FieldSchema(
    ...             field_name="instrumentId",
    ...             display_name="Instrument ID",
    ...             description="The unique identifier",
    ...             data_type_id=datatype.ResourceId(scope="system", code="string"),
    ...             required=True,
    ...             category="SeriesIdentifier"
    ...         )
    ...     ]
    ... )

    Attributes
    ----------
    id : str
        Resource identifier.
    scope : str
        Scope of the dataset definition.
    code : str
        Code of the dataset definition.
    display_name : str
        Human-readable name for the dataset.
    description : str | None
        Optional description of the dataset.
    applicable_entity_types : List[str]
        List of entity types this dataset can be applied to.
    field_schema : List[FieldSchema]
        The schema defining the fields in this dataset.
    """

    id: str = Field(exclude=True)
    scope: str
    code: str
    display_name: str
    description: str | None = None
    applicable_entity_types: List[ApplicableEntityType]
    field_schema: List[FieldSchema]

    @model_validator(mode="before")
    @classmethod
    def des_model(cls, data: Any, info) -> dict[str, Any]:
        if not isinstance(data, dict):
            return data
        if info.context and info.context.get("id"):
            return data | {"id": info.context.get("id")}
        return data

    @computed_field(alias="id")
    def dataset_id(self) -> dict[str, str]:
        return {"scope": self.scope, "code": self.code}

    def read(self, client: httpxClient, old_state: SimpleNamespace):
        scope = old_state.scope
        code = old_state.code
        return client.get(f"/api/api/relationaldatasetdefinitions/{scope}/{code}").json()

    def create(self, client: httpxClient):
        desired = self.model_dump(
            mode="json",
            exclude_none=True,
            by_alias=True,
            exclude={"scope", "code"}
        )
        client.post("/api/api/relationaldatasetdefinitions", json=desired)
        return {
            "scope": self.scope,
            "code": self.code,
        }

    def update(self, client: httpxClient, old_state: SimpleNamespace) -> dict[str, Any] | None:
        if [self.scope, self.code] != [old_state.scope, old_state.code]:
            self.delete(client, old_state)
            return self.create(client)

        remote = self.read(client, old_state)

        # Remove fields that shouldn't be compared
        remote.pop("version", None)
        remote.pop("href", None)
        remote.pop("links", None)
        remote.pop("id", None)

        # Normalize fieldSchema - remove None values from remote to match desired
        # This is required as | is a shallow merge
        if "fieldSchema" in remote:
            for field in remote["fieldSchema"]:
                # Remove None/null values that won't be in desired (due to exclude_none=True)
                if field.get("displayName") is None:
                    field.pop("displayName", None)
                if field.get("description") is None:
                    field.pop("description", None)

        desired = self.model_dump(
            mode="json",
            exclude_none=True,
            by_alias=True,
            exclude={"dataset_id", "scope", "code"}
        )

        effective = remote | desired

        # If updated values are the same as what we currently have from API call
        # then we know we have nothing to update
        if effective == remote:
            return None

        client.put(
            f"/api/api/relationaldatasetdefinitions/{self.scope}/{self.code}",
            json=desired
        )
        return {
            "scope": self.scope,
            "code": self.code,
        }

    @staticmethod
    def delete(client: httpxClient, old_state: SimpleNamespace) -> None:
        client.delete(
            f"/api/api/relationaldatasetdefinitions/{old_state.scope}/{old_state.code}"
        )

    @classmethod
    def create_resource_id(cls, data: dict) -> str:
        return f"relation-dataset-definition_{data['scope']}_{data['code']}".lower()

    @classmethod
    def list_endpoint(cls, **kwargs) -> str:
        return "/api/api/relationaldatasetdefinitions"

    @classmethod
    def fetch_endpoint(cls) -> str:
        return "/api/api/relationaldatasetdefinitions/{scope}/{code}"

    @classmethod
    def from_api(cls, data: dict[str, Any]) -> dict[str, Any]:
        data = data | data["id"]
        data.pop("href", None)
        data.pop("version", None)
        data.pop("links", None)
        data["id"] = cls.create_resource_id(data["id"])

        return data

    @classmethod
    def get_keys(cls, data: dict[str, Any]) -> dict[str, Any]:
        return data["id"]

    def deps(self) -> list[Resource | Ref]:
        # Collect any DataType dependencies from field schemas
        deps = []
        for field in self.field_schema:
            if isinstance(field.data_type_id, (Resource, Ref)):
                deps.append(field.data_type_id)
        return deps


# Serialization helpers for use as keys in other resources
def ser_dataset_key(value, info):
    if info.context and info.context.get("style") == "dump":
        return {"$ref": value.id}
    return {"scope": value.scope, "code": value.code}


def des_dataset_key(value, info):
    if not isinstance(value, dict):
        return value
    if info.context and info.context.get("$refs"):
        ref = info.context["$refs"][value["$ref"]]
        return ref
    return value


RelationalDatasetDefinitionKey = Annotated[
    RelationalDatasetDefinitionResource | RelationalDatasetDefinitionRef,
    BeforeValidator(des_dataset_key),
    PlainSerializer(ser_dataset_key),
    WithJsonSchema(
        {
            "type": "object",
            "properties": {"$ref": {"type": "string", "format": "Key.RelationalDatasetDefinition"}},
            "required": ["$ref"],
        }
    ),
]



from fbnconfig import Deployment, datatype, property

"""
An example configuration for defining property related entities.
The script configures the following entities:
- Property definition


More information can be found here:
https://support.lusid.com/knowledgebase/article/KA-01855/
"""


def configure(env):
    dom_ccy = property.DefinitionRef(
        id="dfdomccy", domain=property.Domain.Holding, scope="default", code="DfDomCcy"
    )

    nominal = property.DefinitionRef(
        id="nominal", domain=property.Domain.Holding, scope="default", code="Nominal"
    )

    rating = property.DefinitionResource(
        id="rating",
        domain=property.Domain.Instrument,
        scope="sc1",
        code="rating",
        display_name="Rating",
        data_type_id=datatype.DataTypeRef(id="default_str", scope="system", code="string"),
        constraint_style=property.ConstraintStyle.Collection,
        property_description="Example property representing a rating",
        life_time=property.LifeTime.Perpetual,
        collection_type=property.CollectionType.Array,
    )

    instrument_property_definition = property.DefinitionResource(
        id="pd1",
        domain=property.Domain.Instrument,
        scope="sc1",
        code="pd1",
        display_name="Property definition example",
        data_type_id=property.ResourceId(scope="system", code="number"),
        constraint_style=property.ConstraintStyle.Property,
        property_description="Example property definition",
        life_time=property.LifeTime.Perpetual,
        collection_type=None,
    )

    pv_nominal = property.DefinitionResource(
        id="derived",
        domain=property.Domain.Holding,
        data_type_id=property.ResourceId(scope="system", code="number"),
        scope="sc1",
        code="PVNominal",
        property_description="Example derived property",
        display_name="DF Nominal",
        derivation_formula=property.Formula("{x} * {y}", x=dom_ccy, y=nominal),
    )

    derived_property = property.DefinitionResource(
        id="derived_property",
        domain=property.Domain.Holding,
        data_type_id=property.ResourceId(scope="system", code="number"),
        scope="sc1",
        code="derived_property",
        property_description="pd1 x df x nominal",
        display_name="DF Nominal pd1",
        derivation_formula=property.Formula(
            "{x} * {y}",
            x=pv_nominal,
            y=instrument_property_definition
        ),
    )

    return Deployment(
        "property_example",
        [instrument_property_definition, dom_ccy, nominal, pv_nominal, derived_property, rating],
    )



from __future__ import annotations

import hashlib
import json
from datetime import datetime
from enum import StrEnum
from types import SimpleNamespace
from typing import Annotated, Any, Literal, Sequence

import httpx
from httpx import Client as httpxClient
from pydantic import (
    BaseModel,
    BeforeValidator,
    ConfigDict,
    Discriminator,
    Field,
    PlainSerializer,
    WithJsonSchema,
    computed_field,
    field_serializer,
    model_serializer,
    model_validator,
)

from . import identity, lumi, scheduler
from .resource_abc import CamelAlias, FetchMixin, ListMixin, Ref, Resource, register_resource

GROUP = "Workflows"


@register_resource(group=GROUP)
class WorkerRef(BaseModel, Ref):
    """Reference an existing worker

    Example
    ----------
    >>> from fbnconfig import workflows
    >>> workflows.WorkerRef(
    ...  id="reconciliation-worker",
    ...  scope="default",
    ...  code="GroupReconciliation")


    Attributes
    ----------
    id : str
         Resource identifier.
    scope : str
        Scope of the worker.
    code: str
        Code of the worker.
    """

    id: str = Field(exclude=True)
    scope: str
    code: str

    def attach(self, client):
        try:
            client.get(f"/workflow/api/workers/{self.scope}/{self.code}")
        except httpx.HTTPStatusError as ex:
            if ex.response.status_code == 404:
                raise RuntimeError(f"Worker {self.scope}/{self.code} does not exist")
            else:
                raise ex


@register_resource(group=GROUP, keys=["scope", "code"])
class WorkerResource(BaseModel, Resource, ListMixin, CamelAlias, FetchMixin):
    """Define a worker

    Example
    ----------
    >>> from fbnconfig import workflows
    >>> workflows.WorkerResource(
    ...  id="myworker",
    ...  scope="default",
    ...  code="MyWorker",
    ...  display_name="DoSomeWork",
    ...  description="New description",
    ...  worker_configuration=workflows.LuminesceView(
    ...    view=lumi.ViewRef(id="my-view", provider="Views.Unit.Something"))
    ... )


    Attributes
    ----------
    id : str
         Resource identifier.
    scope : str
        Scope of the worker.
    code: str
        Code of the worker.
    display_name: str
        Human-readable name for the worker
    description: str | None
        Optional description of the worker
    worker_configuration: Fail | HealthCheck | LuminesceView | SchedulerJob | Sleep
        Worker configuration

    """

    id: str = Field(exclude=True)
    scope: str
    code: str
    display_name: str
    description: str | None = None
    worker_configuration: Annotated[
        Fail | HealthCheck | LuminesceView | SchedulerJob | Sleep, Discriminator("type")
    ]

    @model_validator(mode="before")
    @classmethod
    def des_model(cls, data: Any, info) -> dict[str, Any]:
        if not isinstance(data, dict):
            return data
        if info.context and info.context.get("id"):
            return data | {"id": info.context.get("id")}
        return data

    @computed_field(alias="id")
    def worker_id(self) -> dict[str, str]:
        return {"scope": self.scope, "code": self.code}

    def read(self, client: httpxClient, old_state: SimpleNamespace) -> dict[str, Any] | None:
        scope = old_state.scope
        code = old_state.code
        return client.get(f"/workflow/api/workers/{scope}/{code}").json()

    # noinspection PyProtectedMember
    def __resolve_worker_version(self) -> str:
        if isinstance(self.worker_configuration, SchedulerJob):
            if isinstance(self.worker_configuration.job, scheduler.JobResource):
                dump = self.worker_configuration.job.model_dump(
                    mode="json", exclude_none=True, by_alias=True,
                    exclude={"scope", "code", "image"}
                )
                return hashlib.sha256(json.dumps(dump, sort_keys=True).encode()).hexdigest()
            if isinstance(self.worker_configuration.job, scheduler.JobRef):
                # noinspection PyProtectedMember
                version = self.worker_configuration.job._content_hash
                ref_id = f"{self.worker_configuration.job_id}"
                assert version is not None, f"Expected JobRef {ref_id} to be attached"
                return version

        if isinstance(self.worker_configuration, LuminesceView):
            if isinstance(self.worker_configuration.view, lumi.ViewResource):
                return self.worker_configuration.view._get_content_hash()
            if isinstance(self.worker_configuration.view, lumi.ViewRef):
                return self.worker_configuration.view._version or "0"
        return "0"

    def create(self, client: httpxClient) -> dict[str, Any] | None:
        desired = self.model_dump(mode="json", exclude_none=True,
                                  by_alias=True, exclude={"scope", "code"})
        client.post("/workflow/api/workers", json=desired)

        return {
            "id": self.id,
            "scope": self.scope,
            "code": self.code,
            "worker_version": self.__resolve_worker_version(),
        }

    def update(self, client: httpxClient, old_state: SimpleNamespace) -> dict[str, Any] | None:
        if [self.scope, self.code] != [old_state.scope, old_state.code]:
            self.delete(client, old_state)
            return self.create(client)
        remote = self.read(client, old_state) or {}
        remote.pop("version")
        remote.pop("parameters")
        remote.pop("resultFields")
        remote.pop("links")
        remote.pop("id")
        # dump the api format with scope, code in a resourceId
        desired = self.model_dump(mode="json", exclude_none=True,
                                  by_alias=True, exclude={"worker_id", "scope", "code"})
        effective = remote | desired
        if effective == remote and old_state.worker_version == self.__resolve_worker_version():
            return None

        client.put(f"/workflow/api/workers/{self.scope}/{self.code}", json=desired).json()
        return {
            "id": self.id,
            "scope": self.scope,
            "code": self.code,
            "worker_version": self.__resolve_worker_version(),
        }

    @staticmethod
    def delete(client: httpxClient, old_state: SimpleNamespace) -> None:
        client.delete(f"/workflow/api/workers/{old_state.scope}/{old_state.code}")

    @classmethod
    def create_resource_id(cls, data: dict) -> str:
        return f"worker_{data['scope']}_{data['code']}".lower()

    @classmethod
    def fetch_endpoint(cls) -> str:
        return "/workflow/api/workers/{scope}/{code}"

    @classmethod
    def list_endpoint(cls, **kwargs) -> str:
        return "/workflow/api/workers"

    @classmethod
    def from_api(cls, data: dict[str, Any]) -> dict[str, Any]:
        data.pop("links", None)
        data.pop("version", None)
        data = data | data["id"]
        data["id"] = cls.create_resource_id(data)
        wc = data.get("workerConfiguration")
        if wc and (wc_type := wc.get("type")):
            if wc_type == "LuminesceView":
                name = wc.pop("name", None)
                wc["view"] = {"$ref": f"view_{name}".lower()}
            elif wc_type == "SchedulerJob":
                job_id = wc.pop("jobId", None)
                wc["job"] = {
                    "$ref": f"job_{job_id['scope']}_{job_id['code']}".lower()
                }
        return data

    @classmethod
    def get_keys(cls, data: dict[str, Any]) -> dict[str, Any]:
        return data["id"]

    def deps(self) -> list[Any]:
        if isinstance(self.worker_configuration, LuminesceView) and isinstance(
            self.worker_configuration.view, (lumi.ViewResource, lumi.ViewRef)
        ):
            return [self.worker_configuration.view]
        if isinstance(self.worker_configuration, SchedulerJob):
            return [self.worker_configuration.job]
        return []


def ser_worker_key(value, info):
    if info.context and info.context.get("style") == "dump":
        return {"$ref": value.id}
    return {"scope": value.scope, "code": value.code}


def des_worker_key(value, info):
    if not isinstance(value, dict):
        return value
    if info.context and info.context.get("$refs"):
        ref = info.context["$refs"][value["$ref"]]
        return ref
    return value


WorkerKey = Annotated[
    WorkerResource | WorkerRef,
    BeforeValidator(des_worker_key),
    PlainSerializer(ser_worker_key),
    WithJsonSchema(
        {
            "type": "object",
            "properties": {"$ref": {"type": "string", "format": "Key.Worker"}},
            "required": ["$ref"],
        }
    ),
]


@register_resource(group=GROUP)
class TaskDefinitionRef(BaseModel, Ref):
    """Reference an existing task definition

    Example
    ----------
    >>> from fbnconfig import workflows
    >>> workflows.TaskDefinitionRef(
    ...  id="task-definition-ref",
    ...  scope="DQ",
    ...  code="ExceptionTask")


    Attributes
    ----------
    id : str
         Resource identifier.
    scope : str
        Scope of the task definition.
    code: str
        Code of the task definition.
    """

    id: str = Field(exclude=True)
    scope: str
    code: str

    def attach(self, client):
        try:
            client.get(f"/workflow/api/taskdefinitions/{self.scope}/{self.code}").json()
        except httpx.HTTPStatusError as ex:
            if ex.response.status_code == 404:
                raise RuntimeError(f"Task definition {self.scope}/{self.code} does not exist")
            else:
                raise ex


@register_resource(group=GROUP, keys=["scope", "code"])
class TaskDefinitionResource(BaseModel, Resource, ListMixin, FetchMixin):
    id: str = Field(exclude=True)
    scope: str
    code: str
    display_name: str
    initial_state: InitialState
    states: list[TaskStateDefinition]
    description: str | None = None
    triggers: list[TriggerDefinition] | None = None
    transitions: list[TaskTransitionDefinition] | None = None
    field_schema: list[TaskFieldDefinition] | None = None
    actions: list[ActionDefinition] | None = None

    @model_validator(mode="before")
    @classmethod
    def des_model(cls, data: Any, info) -> dict[str, Any]:
        if not isinstance(data, dict):
            return data
        if info.context and info.context.get("id"):
            return data | {"id": info.context.get("id")}
        return data

    @computed_field(alias="id")
    def task_def_id(self) -> dict[str, str]:
        return {"scope": self.scope, "code": self.code}

    def read(self, client: httpxClient, old_state: SimpleNamespace) -> None | dict[str, Any]:
        remote = client.get(f"/workflow/api/taskdefinitions/{old_state.scope}/{old_state.code}").json()
        return remote

    def create(self, client: httpxClient) -> dict[str, Any] | None:
        desired = self.model_dump(mode="json", by_alias=True, exclude_none=True,
                                  exclude={"scope", "code"})
        source_version = hashlib.sha256(json.dumps(desired, sort_keys=True).encode()).hexdigest()
        remote_version = client.post("/workflow/api/taskdefinitions", json=desired).json()["version"][
            "asAtVersionNumber"
        ]
        return {
            "id": self.id,
            "scope": self.scope,
            "code": self.code,
            "source_version": source_version,
            "remote_version": remote_version,
        }

    def update(self, client: httpxClient, old_state) -> dict[str, Any] | None:
        if self.scope != old_state.scope or self.code != old_state.code:
            self.delete(client, old_state)
            return self.create(client)

        remote = self.read(client, old_state) or {}
        desired = self.model_dump(mode="json", by_alias=True, exclude_none=True,
                                  exclude={"scope", "code"})
        remote_version = remote["version"]["asAtVersionNumber"]
        source_version = hashlib.sha256(json.dumps(desired, sort_keys=True).encode()).hexdigest()
        if source_version == old_state.source_version and remote_version == old_state.remote_version:
            return None

        remote_version = client.put(
            f"/workflow/api/taskdefinitions/{self.scope}/{self.code}", json=desired
        ).json()["version"]["asAtVersionNumber"]
        return {
            "id": self.id,
            "scope": self.scope,
            "code": self.code,
            "source_version": source_version,
            "remote_version": remote_version,
        }

    @staticmethod
    def delete(client: httpxClient, old_state) -> None:
        client.delete(f"/workflow/api/taskdefinitions/{old_state.scope}/{old_state.code}")

    @classmethod
    def list_endpoint(cls, **kwargs) -> str:
        return "/workflow/api/taskdefinitions"

    @classmethod
    def create_resource_id(cls, data: dict) -> str:
        return f"task-definition_{data['scope']}_{data['code']}".lower()

    @classmethod
    def get_keys(cls, data: dict[str, Any]) -> dict[str, Any]:
        return data["id"]

    @classmethod
    def fetch_endpoint(cls) -> str:
        return "/workflow/api/taskdefinitions/{scope}/{code}"

    @classmethod
    def from_api(cls, data: dict[str, Any]) -> dict[str, Any]:
        data = data | data["id"]
        res_id = cls.create_resource_id(data)
        result = {
            **data,
            "id": res_id,
        }
        result.pop("version", None)

        for state in result.get("states", []):
            trigger = state.get("trigger")
            if trigger and "type" in trigger and "name" not in trigger:
                state["trigger"] = {
                    "name": state["name"],
                    "type": trigger["type"],
                }

        if "triggers" in result:
            normalized_triggers = []
            for t in result["triggers"]:
                if "trigger" in t:
                    normalized_triggers.append({
                        "name": t["name"],
                        "type": t["trigger"]["type"],
                    })
                else:
                    normalized_triggers.append(t)
            result["triggers"] = normalized_triggers

        return result

    def deps(self) -> list[Any]:
        deps = []
        for t in self.transitions if self.transitions else []:
            deps.extend(t.deps())
        for a in self.actions if self.actions else []:
            deps.extend(a.deps())

        return deps


def ser_taskdefinition_key(value, info):
    if info.context and info.context.get("style") == "dump":
        return {"$ref": value.id}
    return {"scope": value.scope, "code": value.code}


def des_taskdefinition_key(value, info):
    if not isinstance(value, dict):
        return value
    if info.context and info.context.get("$refs") and "$ref" in value:
        ref = info.context["$refs"].get(value["$ref"])
        if ref:
            return ref
    return value


TaskDefinitionKey = Annotated[
    TaskDefinitionResource | TaskDefinitionRef,
    BeforeValidator(des_taskdefinition_key),
    PlainSerializer(ser_taskdefinition_key),
    WithJsonSchema(
        {
            "type": "object",
            "properties": {"$ref": {"type": "string", "format": "Key.TaskDefinition"}},
            "required": ["$ref"],
        }
    ),
]


class EventStatus(StrEnum):
    ACTIVE = "Active"
    INACTIVE = "Inactive"


class EventMatchingPattern(BaseModel, CamelAlias):
    event_type: str
    filter: str | None = None


class CalendarReference(BaseModel, CamelAlias):
    scope: str
    code: str


class ScheduleMatchingPatternContext(BaseModel, CamelAlias):
    time_zone: str
    holiday_calendars: Sequence[CalendarReference] | None = None


class SpecifiedTime(BaseModel, CamelAlias):
    type: Literal["Specified"] = Field("Specified", init=False)
    hours: int
    minutes: int


class CutLabelReference(BaseModel, CamelAlias):
    type: Literal["CutLabelReference"] = Field("CutLabelReference", init=False)
    code: str


TimesOfDay = Annotated[SpecifiedTime | CutLabelReference, Field(discriminator="type")]


class TimeConstraints(BaseModel, CamelAlias):
    start_date: str
    end_date: str | None = None
    times_of_day: Sequence[SpecifiedTime | CutLabelReference]


class RelativeMonthRegularity(BaseModel, CamelAlias):
    type: Literal["RelativeMonth"] = Field("RelativeMonth", init=False)
    frequency: int
    days_of_week: Sequence[str]
    index: str


class SpecificMonthRegularity(BaseModel, CamelAlias):
    type: Literal["SpecificMonth"] = Field("SpecificMonth", init=False)
    frequency: int
    days_of_month: Sequence[int]


class WeekRegularity(BaseModel, CamelAlias):
    type: Literal["Week"] = Field("Week", init=False)
    frequency: int
    days_of_week: Sequence[str]


class DayOfYear(BaseModel, CamelAlias):
    month: int
    day: int


class YearRegularity(BaseModel, CamelAlias):
    type: Literal["Year"] = Field("Year", init=False)
    dates: Sequence[DayOfYear]


class DayRegularity(BaseModel, CamelAlias):
    type: Literal["Day"] = Field("Day", init=False)
    frequency: int


DateRegularity = Annotated[
    RelativeMonthRegularity | SpecificMonthRegularity | WeekRegularity | YearRegularity | DayRegularity,
    Field(discriminator="type")
]


class RecurrencePattern(BaseModel, CamelAlias):
    time_constraints: TimeConstraints
    business_day_adjustment: str
    date_regularity: DateRegularity


class ScheduleMatchingPattern(BaseModel, CamelAlias):
    context: ScheduleMatchingPatternContext
    recurrence_pattern: RecurrencePattern


class EventHandlerMapping(BaseModel, CamelAlias):
    map_from: str | None = None
    set_to: str | None | identity.UserKey = None


class CreateNewTaskActivity(BaseModel, CamelAlias):
    type: Literal["CreateNewTask"] = Field("CreateNewTask", init=False)
    correlation_ids: list[EventHandlerMapping] | None = None
    task_fields: dict[str | TaskFieldDefinition, FieldMapping] | None = None
    initial_trigger: str | TriggerDefinition | None = None

    @field_serializer("initial_trigger")
    def trigger_serializer(self, initial_trigger: str | TriggerDefinition | None) -> str | None:
        return (
            initial_trigger.name if isinstance(initial_trigger, TriggerDefinition) else initial_trigger
        )

    @field_serializer("task_fields")
    def task_fields_serializer(self, task_fields: dict[str | TaskFieldDefinition, FieldMapping]):
        return {k.name if isinstance(k, TaskFieldDefinition) else k: v for k, v in task_fields.items()}


class UpdateMatchingTasksActivity(BaseModel, CamelAlias):
    type: Literal["UpdateMatchingTasks"] = Field("UpdateMatchingTasks", init=False)
    correlation_ids: list[EventHandlerMapping] | None = None
    task_fields: dict[str, FieldMapping] | None = None
    filter: str | None = None
    trigger: str | TriggerDefinition

    @field_serializer("trigger")
    def trigger_serializer(self, trigger: str | TriggerDefinition) -> str:
        return trigger.name if isinstance(trigger, TriggerDefinition) else trigger


class Fail(BaseModel):
    type: Literal["Fail"] = Field("Fail", init=False)


class HealthCheck(BaseModel):
    type: Literal["HealthCheck"] = Field("HealthCheck", init=False)
    url: str


class LuminesceView(BaseModel):
    type: Literal["LuminesceView"] = Field("LuminesceView", init=False)
    view: lumi.ViewKey = Field(exclude=True, init=True)

    @computed_field
    def name(self) -> str:
        return self.view.provider


class SchedulerJob(BaseModel):
    type: Literal["SchedulerJob"] = Field("SchedulerJob", init=False)
    job: scheduler.JobRef | scheduler.JobResource = Field(exclude=True, init=True)

    @computed_field(alias="jobId")
    def job_id(self) -> dict[str, str]:
        assert self.job is not None, "Job should not be None"
        return {"scope": self.job.scope, "code": self.job.code}


class Sleep(BaseModel):
    type: Literal["Sleep"] = Field("Sleep", init=False)


class TaskStateDefinition(BaseModel):
    model_config = ConfigDict(use_enum_values=True, frozen=True)
    name: str


class TaskFieldDefinitionType(StrEnum):
    STRING = "String"
    DECIMAL = "Decimal"
    DATETIME = "DateTime"
    BOOLEAN = "Boolean"


class ReadOnlyStateType(StrEnum):
    ALL_STATES = "AllStates"
    INITIAL_STATE = "InitialState"
    TERMINAL_STATE = "TerminalState"
    SELECTED_STATES = "SelectedStates"


class ReadOnlyStates(BaseModel, CamelAlias):
    model_config = ConfigDict(use_enum_values=True, frozen=True)
    state_type: str | ReadOnlyStateType
    selected_states: list[str | TaskStateDefinition] | None = None

    @field_serializer("selected_states")
    def states(self, selected_states) -> list[str] | None:
        return [s.name if isinstance(s, TaskStateDefinition) else s for s in selected_states]

    def __hash__(self) -> int:
        return hash((self.state_type, frozenset(self.selected_states or frozenset())))


class ValueConstraintType(StrEnum):
    SUGGESTED = "Suggested"
    VALIDATED = "Validated"


class ValueSourceType(StrEnum):
    ACCEPTABLE_VALUES = "AcceptableValues"


class ValueConstraints(BaseModel, CamelAlias):
    model_config = ConfigDict(use_enum_values=True, frozen=True)
    constraint_type: str | ValueConstraintType
    value_source_type: str | ValueSourceType
    acceptable_values: list[str]

    def __hash__(self) -> int:
        return hash((self.constraint_type, self.value_source_type, frozenset(self.acceptable_values)))


class TaskFieldDefinition(BaseModel, CamelAlias):
    model_config = ConfigDict(use_enum_values=True, frozen=True)
    name: str
    type: str | TaskFieldDefinitionType
    read_only_states: ReadOnlyStates | None = None
    value_constraints: ValueConstraints | None = None

    @field_serializer("type")
    def type_serializer(self, field_type: str | TaskFieldDefinitionType | None):
        return field_type.name if isinstance(field_type, TaskFieldDefinitionType) else field_type


class InitialState(BaseModel, CamelAlias):
    name: str | TaskStateDefinition
    required_fields: list[str | TaskFieldDefinition] | None = []

    @field_serializer("name")
    def name_serializer(self, name: str | TaskStateDefinition):
        return name.name if isinstance(name, TaskStateDefinition) else name

    @field_serializer("required_fields")
    def req_fields(self, required_fields) -> list[str] | None:
        return [f.name if isinstance(f, TaskFieldDefinition) else f for f in required_fields]


class TriggerDefinition(BaseModel):
    name: str
    type: str

    @model_serializer()
    def serialize_model(self):
        return {"name": self.name, "trigger": {"type": self.type}}


class FieldMapping(BaseModel, CamelAlias):
    map_from: str | TaskFieldDefinition | None = None
    set_to: Any = None

    @field_serializer("map_from")
    def map_serializer(self, value: str | TaskFieldDefinition | None) -> Any | None:
        return value.name if isinstance(value, TaskFieldDefinition) else value


class ChildTaskConfiguration(BaseModel, CamelAlias):
    task_definition: TaskDefinitionKey = Field(serialization_alias="taskDefinitionId")
    task_definition_as_at: str | datetime | None = None
    initial_trigger: str | TriggerDefinition | None = None
    child_task_fields: dict[str | TaskFieldDefinition, FieldMapping] | None = None
    map_stacking_key_from: str | TaskFieldDefinition | None = None

    @field_serializer("task_definition_as_at")
    def task_def_as_at(self, task_as_at: str | datetime | None) -> str | None:
        return task_as_at.isoformat() if isinstance(task_as_at, datetime) else task_as_at

    @field_serializer("initial_trigger")
    def trigger(self, initial_trigger: str | TriggerDefinition | None) -> str | None:
        return (
            initial_trigger.name if isinstance(initial_trigger, TriggerDefinition) else initial_trigger
        )

    @field_serializer("map_stacking_key_from")
    def mskf(self, map_stacking_key_from: str | TaskFieldDefinition | None) -> str | None:
        return (
            map_stacking_key_from.name
            if isinstance(map_stacking_key_from, TaskFieldDefinition)
            else map_stacking_key_from
        )

    @field_serializer("child_task_fields")
    def ctf(self, child_task_fields: dict[str | TaskFieldDefinition, FieldMapping]):
        return {
            k.name if isinstance(k, TaskFieldDefinition) else k: v for k, v in child_task_fields.items()
        }


class CreateChildTasksAction(BaseModel, CamelAlias):
    type: str = Field(init=False, default="CreateChildTasks")
    child_task_configurations: list[ChildTaskConfiguration]

    def deps(self) -> list[TaskDefinitionResource | TaskDefinitionRef]:
        return [c.task_definition for c in self.child_task_configurations]


class WorkerStatusTriggers(BaseModel, CamelAlias):
    started: str | TriggerDefinition | None = None
    completed_with_results: str | TriggerDefinition | None = None
    completed_no_results: str | TriggerDefinition | None = None
    failed_to_start: str | TriggerDefinition | None = None
    failed_to_complete: str | TriggerDefinition | None = None

    @field_serializer("*")
    def serialize_fields(self, value) -> str:
        return value.name if isinstance(value, TriggerDefinition) else value


class TriggerParentTaskAction(BaseModel):
    type: str = Field(default="TriggerParentTask", init=False)
    trigger: str | TriggerDefinition

    @field_serializer("trigger")
    def trigger_serialization(self, trigger: str | TriggerDefinition) -> str:
        return trigger.name if isinstance(trigger, TriggerDefinition) else trigger


class ResultantChildTaskConfiguration(BaseModel, CamelAlias):
    child_task_configuration: ChildTaskConfiguration
    result_matching_pattern: str | None = None

    @model_serializer()
    def serialize_model(self):
        result_pattern = (
            {"resultMatchingPattern": {"filter": self.result_matching_pattern}}
            if self.result_matching_pattern
            else None
        )
        if result_pattern is None:
            return self.child_task_configuration.model_dump(by_alias=True, exclude_none=True)
        else:
            return (
                self.child_task_configuration.model_dump(by_alias=True, exclude_none=True)
                | result_pattern
            )


class RunWorkerAction(BaseModel, CamelAlias):
    type: str = Field(init=False, default="RunWorker")
    worker: WorkerKey = Field(exclude=True)  # sent as workerId
    worker_as_at: str | datetime | None = None
    worker_parameters: dict[str | TaskFieldDefinition, FieldMapping] | None = None
    worker_status_triggers: WorkerStatusTriggers | None = None
    child_task_configurations: list[ResultantChildTaskConfiguration] | None = None
    # in seconds
    worker_timeout: int | None = None

    @computed_field()
    def worker_id(self) -> WorkerKey:
        return self.worker

    @field_serializer("worker_as_at")
    def worker_as_at_s(self, worker_as_at: str | datetime | None) -> str | None:
        return worker_as_at.isoformat() if isinstance(worker_as_at, datetime) else worker_as_at

    @field_serializer("worker_parameters")
    def worker_params(self, worker_parameters: dict[str | TaskFieldDefinition, FieldMapping]):
        return {
            k.name if isinstance(k, TaskFieldDefinition) else k: v for k, v in worker_parameters.items()
        }

    def deps(self) -> list[Any]:
        child_tasks = self.child_task_configurations if self.child_task_configurations else []
        return [self.worker] + [c.child_task_configuration.task_definition for c in child_tasks]


class ActionDefinition(BaseModel, CamelAlias):
    name: str
    action_details: CreateChildTasksAction | RunWorkerAction | TriggerParentTaskAction
    run_as_user_id: str | identity.UserKey | None = None

    def deps(self) -> list[Any]:
        result = []
        if isinstance(self.run_as_user_id, (identity.UserResource, identity.UserRef)):
            result.append(self.run_as_user_id)
        if isinstance(self.action_details, (CreateChildTasksAction, RunWorkerAction)):
            result.extend(self.action_details.deps())
        return result


class TaskTransitionDefinition(BaseModel, CamelAlias):
    from_state: str | TaskStateDefinition
    to_state: str | TaskStateDefinition
    trigger: str | TriggerDefinition
    guard: str | None = None
    action: str | ActionDefinition | None = None

    @field_serializer("from_state", "to_state", "trigger", "action")
    def state_serializer(
        self, value: str | TaskStateDefinition | TriggerDefinition | ActionDefinition
    ) -> str:
        return (
            value.name
            if isinstance(value, (TaskStateDefinition, TriggerDefinition, ActionDefinition))
            else value
        )

    def deps(self):
        return self.action.deps() if isinstance(self.action, ActionDefinition) else []


@register_resource(group=GROUP)
class EventHandlerRef(BaseModel, Ref):
    """Reference an existing event handler

    Example
    ----------
    >>> from fbnconfig import workflows
    >>> workflows.EventHandlerRef(
    ...  id="reconciliation-event-handler",
    ...  scope="reconciliation",
    ...  code="start-event")
    """

    id: str = Field(exclude=True)
    scope: str
    code: str

    def attach(self, client) -> None:
        try:
            client.get(f"/workflow/api/eventhandlers/{self.scope}/{self.code}")
        except httpx.HTTPStatusError as ex:
            if ex.response.status_code == 404:
                raise RuntimeError(f"Event handler {self.scope}/{self.code} does not exist")
            else:
                raise ex


@register_resource(group=GROUP, keys=["scope", "code"])
class EventHandlerResource(BaseModel, Resource, ListMixin, CamelAlias, FetchMixin):
    """
    Create an event handler

    Example
    -------
    >>> from fbnconfig import workflows
    >>> event_handler = EventHandlerResource(
    >>> id="event_handler_id",
    >>> scope="scope",
    >>> code="code",
    >>> display_name="event_handler_name",
    >>> description="event_handler_description",
    >>> status="initial_status",
    >>> event_matching_pattern=EventMatchingPattern(event_type="AllocationCreated",filter=None),
    >>> run_as_user_id=EventHandlerMapping(set_to="ExampleUser"),
    >>> task_definition_id=TaskDefinitionRef(
    >>>         id="task_definition_id",
    >>>         scope="scope",
    >>>         code="task_def_code"),
    >>> task_activity=CreateNewTaskActivity(type="CreateNewTask", initial_trigger="Start")
    >>> )
    """

    id: str = Field(exclude=True)
    scope: str
    code: str
    display_name: str
    description: str | None = None
    status: str | EventStatus
    event_matching_pattern: EventMatchingPattern | None = None
    run_as_user_id: EventHandlerMapping | None = None
    task_definition: TaskDefinitionKey = Field(
        validation_alias="taskDefinitionId", serialization_alias="taskDefinitionId"
    )
    task_definition_as_at: datetime | None = None
    task_activity: Annotated[
        CreateNewTaskActivity | UpdateMatchingTasksActivity, Field(discriminator="type")
    ]
    schedule_matching_pattern: ScheduleMatchingPattern | None = None

    @model_validator(mode="before")
    @classmethod
    def des_model(cls, data: Any, info) -> dict[str, Any]:
        if not isinstance(data, dict):
            return data
        if info.context and info.context.get("id"):
            return data | {"id": info.context.get("id")}
        return data

    @computed_field(alias="id")
    def event_handler_id(self) -> dict[str, str]:
        return {"scope": self.scope, "code": self.code}

    def __get_content_hash__(self) -> str:
        dump = self.model_dump(
            mode="json", exclude_none=True, by_alias=True, exclude={"event_handler_id", "scope", "code"}
        )
        return hashlib.sha256(json.dumps(dump, sort_keys=True).encode()).hexdigest()

    def read(self, client: httpxClient, old_state: SimpleNamespace) -> None | dict[str, Any]:
        scope = old_state.scope
        code = old_state.code
        return client.get(f"/workflow/api/eventhandlers/{scope}/{code}").json()

    def create(self, client: httpxClient) -> dict[str, Any] | None:
        desired = self.model_dump(mode="json", exclude_none=True, by_alias=True,
                                  exclude={"scope", "code"})
        remote = client.post("/workflow/api/eventhandlers", json=desired).json()
        return {
            "id": self.id,
            "scope": self.scope,
            "code": self.code,
            "source_version": self.__get_content_hash__(),
            "remote_version": f"{remote['version']['asAtVersionNumber']}",
        }

    def update(self, client: httpxClient, old_state) -> dict[str, Any] | None:
        if [self.scope, self.code] != [old_state.scope, old_state.code]:
            self.delete(client, old_state)
            return self.create(client)
        remote = self.read(client, old_state) or {}
        remote_version = f"{remote['version']['asAtVersionNumber']}"
        # dump api compatible with scope, code in a resourceId
        desired = self.model_dump(
            mode="json", exclude_none=True, by_alias=True, exclude={"event_handler_id", "scope", "code"}
        )

        source_version = self.__get_content_hash__()

        if source_version == old_state.source_version and remote_version == old_state.remote_version:
            return None

        updated = client.put(
            f"/workflow/api/eventhandlers/{self.scope}/{self.code}", json=desired
        ).json()

        return {
            "id": self.id,
            "scope": self.scope,
            "code": self.code,
            "source_version": source_version,
            "remote_version": f"{updated['version']['asAtVersionNumber']}",
        }

    @staticmethod
    def delete(client: httpxClient, old_state) -> None:
        client.delete(f"/workflow/api/eventhandlers/{old_state.scope}/{old_state.code}")

    @classmethod
    def get_keys(cls, data: dict[str, Any]) -> dict[str, Any]:
        return data["id"]

    @classmethod
    def fetch_endpoint(cls) -> str:
        return "/workflow/api/eventhandlers/{scope}/{code}"

    @classmethod
    def from_api(cls, data: dict[str, Any]) -> dict[str, Any]:
        data = data | data["id"]
        result = {
            **data,
            "id": cls.create_resource_id(data),
        }
        result.pop("version", None)
        td = result.get("taskDefinitionId")
        if isinstance(td, dict):
            result["taskDefinitionId"] = {
                "$ref": TaskDefinitionResource.create_resource_id(
                    {
                        "scope": td["scope"],
                        "code": td["code"]
                    }
                )
            }

        smp = result.get("scheduleMatchingPattern", {})
        rp = smp.get("recurrencePattern", {})
        tc = rp.get("timeConstraints", {})
        times_of_day = tc.get("timesOfDay", [])

        for item in times_of_day:
            if item.get("type") == "CutLabel":
                item["type"] = "CutLabelReference"

        return result

    @classmethod
    def create_resource_id(cls, data: dict) -> str:
        return f"event-handler_{data['scope']}_{data['code']}".lower()

    @classmethod
    def list_endpoint(cls, **kwargs) -> str:
        return "/workflow/api/eventhandlers"

    def deps(self) -> list[Resource | Ref]:
        res = []
        res.append(self.task_definition)
        user_types = (identity.UserResource, identity.UserRef, identity.CurrentUserRef)
        if self.run_as_user_id and isinstance(self.run_as_user_id.set_to, user_types):
            res.append(self.run_as_user_id.set_to)
        return res



import http
import itertools
import json
import logging
import os
import pathlib
import queue
import threading
import time
import uuid
from datetime import datetime, timezone
from typing import Any, AsyncIterator, cast

from anyio import to_thread
from httpx import HTTPStatusError
from pydantic import ValidationError
from starlette.applications import Starlette
from starlette.exceptions import HTTPException
from starlette.middleware import Middleware
from starlette.middleware.base import BaseHTTPMiddleware
from starlette.requests import Request
from starlette.responses import JSONResponse, StreamingResponse
from starlette.routing import Mount, Route

import fbnconfig
from fbnconfig import log as deployment_log
from fbnconfig import schemagen
from fbnconfig.deploy import run as deploy_run
from fbnconfig.deploy import undump_deployment
from fbnconfig.exceptions import ApiError, ApiStatusError, NotFoundError, QueryError
from fbnconfig.load_module import load_module
from server.api import get_fetch, get_list, get_metadata
from server.deploy_stream import DeployStreamRequest, make_event, redact_event, to_jsonl_line

logger = logging.getLogger("fbnconfig.server")  # creates logger for the server


def _get_request_ids(request: Request) -> tuple[str, str | None]:
    """Return (request_id, correlation_id).

    Uses LUSID Web API header convention: Correlationid.
    """
    correlation_id = request.headers.get("Correlationid")
    # Mimic LUSID's pageInstance style: 128-bit ID as 32 hex chars with hyphens.
    request_id = f"page:fbnconfig-api,pageInstance:{uuid.uuid4()}"
    return request_id, correlation_id


def _configure_logging(level_name: str = "INFO") -> None:
    level = logging.getLevelNamesMapping().get(level_name.upper().strip(), logging.INFO)
    root_logger = logging.getLogger()
    # Checking if handlers is none, to avoid adding duplicate logging handlers
    if not root_logger.handlers:
        logging.basicConfig(level=level)
    else:
        root_logger.setLevel(level)


# Based on environment/need, LOG_LEVEL may be different.
_configure_logging(os.getenv("LOG_LEVEL", "INFO"))
# Disable redundant uvicorn access logs (Structured json logs do it anyway)
logging.getLogger("uvicorn.access").disabled = True


def _json_log(level: int, event: str, **fields):
    log_data = {"timestamp": datetime.now(timezone.utc).isoformat(), "event": event, **fields}
    logger.log(level, json.dumps(log_data))


def get_examples_path():
    return pathlib.Path(__file__).parent.parent / "public_examples" / "examples"


async def homepage(_: Request):
    return JSONResponse({"message": "Hello, Banana!", "status": "success"})


async def health_check(_: Request):
    return JSONResponse({"status": "healthy", "service": "fbnconfig-api"})


async def ready_check(_: Request):
    # Returns 503 until startup event completes, then 200
    started_at = getattr(app.state, "started_at", None)
    if started_at is None:
        return JSONResponse({"status": "not_ready", "service": "fbnconfig-api"}, status_code=503)
    return JSONResponse({
        "status": "ready", "service": "fbnconfig-api", "started_at": started_at.isoformat()
    })


async def get_schema(_):
    return JSONResponse(SCHEMA)


async def invoke(request: Request) -> JSONResponse:
    # Minimal callable endpoint to validate POST routing, JSON parsing, Middleware and Error logging.
    try:
        payload = await request.json()
    except Exception as exc:
        _json_log(
            logging.WARNING,
            "json_parse_failed",
            request_id=getattr(request.state, "request_id", None),
            correlation_id=getattr(request.state, "correlation_id", None),
            error=str(exc),
        )
        return JSONResponse(
            {"status": "error", "detail": "Invalid JSON"},
            status_code=400
        )
    return JSONResponse({"status": "ok", "received": payload})


async def list_examples(request: Request) -> JSONResponse:
    example_folder = get_examples_path()
    files = example_folder.glob("*.py")
    res = [
        {"name": str(f.stem), "path": request.url_for("get_example", example_name=str(f.stem)).path}
        for f in files
    ]
    return JSONResponse(res)


async def get_example(request: Request):
    example_folder = get_examples_path()
    example_name = request.path_params["example_name"] + ".py"
    script_path = example_folder / example_name
    module = load_module(script_path, str(example_folder))
    if getattr(module, "configure", None) is None:
        raise HTTPException(status_code=400, detail="No configure found in " + example_name)
    deployment = module.configure({})
    dump = fbnconfig.dump_deployment(deployment)
    return JSONResponse(dump)


def create_client(lusid_env, token):
    if lusid_env is None or token is None:
        raise HTTPException(status_code=401, detail="No auth header or no X-LUSID-Host")
    client = fbnconfig.create_client(lusid_env, token)
    return client


async def get_log(request: Request):
    client = create_client(request.state.lusid_env, request.state.token)
    deployment_name = request.path_params["deployment_name"]
    log = []
    for line in deployment_log.list_resources_for_deployment(client, deployment_id=deployment_name):
        d = line._asdict()
        d["state"] = vars(d["state"])
        log.append(d)
    return JSONResponse(log)


# ---------------------------------------------------------------------------
# POST /deploy/stream — JSONL streaming deployment endpoint
# ---------------------------------------------------------------------------

# Sentinel value placed on the queue to signal "the worker is done."
# Using None as a sentinel is a common pattern — when the reader sees
# None it knows no more items are coming.
_STREAM_SENTINEL = None

_active_workers: list[threading.Thread] = []
_active_workers_lock = threading.Lock()
_WORKER_DRAIN_TIMEOUT_S = 30


class _DeployCancelledError(Exception):
    """Raised inside the worker when the client disconnects mid-stream."""


def _map_exception(exc: Exception) -> dict:
    """Map an exception to a structured error payload for the JSONL stream.

    We classify known exceptions and return safe, informative messages.
    Unknown exceptions get a generic message — never leak stack traces.
    """
    if isinstance(exc, HTTPStatusError):
        return {
            "errorType": "HTTPStatusError",
            "message": str(exc),
            "httpStatus": exc.response.status_code,
        }
    if isinstance(exc, ApiStatusError):
        return {
            "errorType": "ApiStatusError",
            "message": str(exc),
            "httpStatus": exc.http_exception.response.status_code,
        }
    if isinstance(exc, QueryError):
        return {"errorType": "QueryError", "message": str(exc)}
    if isinstance(exc, NotFoundError):
        return {"errorType": "NotFoundError", "message": exc.message}
    if isinstance(exc, ApiError):
        return {"errorType": "ApiError", "message": str(exc)}
    if isinstance(exc, RuntimeError):
        return {"errorType": "RuntimeError", "message": str(exc)}
    if isinstance(exc, ValidationError):
        return {"errorType": "ValidationError", "message": str(exc)}
    if isinstance(exc, KeyError):
        return {"errorType": "KeyError", "message": str(exc)}
    # Generic fallback — don't expose internals
    return {"errorType": "InternalError", "message": "An internal error occurred"}


async def deploy_stream(request: Request):
    """Stream a deployment as JSONL events.

    Flow:
    1. Validate auth + request body (errors → normal HTTP status codes)
    2. Start a worker thread that runs the synchronous deploy.run()
    3. Return a StreamingResponse that yields JSONL lines from a queue
    """

    # --- Auth check (same pattern as other endpoints) ---
    if request.state.token is None or request.state.lusid_env is None:
        return JSONResponse(
            {"detail": "No auth header or no X-LUSID-Host"},
            status_code=401,
        )

    # --- Build the set of secrets to redact from streamed output ---
    secrets: set[str] = set()
    if request.state.token:
        secrets.add(request.state.token)
    vso_header = request.headers.get("x-fbn-verified-service-origin")
    if vso_header:
        secrets.add(vso_header)

    # --- Parse & validate request body ---
    try:
        body = await request.json()
    except Exception:
        return JSONResponse({"detail": "Invalid JSON"}, status_code=400)

    try:
        req = DeployStreamRequest.model_validate(body)
    except ValidationError as exc:
        return JSONResponse({"detail": exc.errors()}, status_code=422)

    # --- Extract values we'll use throughout the stream ---
    deployment_id = req.deployment.deployment_id
    request_id = request.state.request_id
    correlation_id = request.state.correlation_id

    _json_log(
        logging.INFO,
        "deploy_stream_started",
        request_id=request_id,
        correlation_id=correlation_id,
        deployment_id=deployment_id,
    )
    stream_start = time.perf_counter()

    # --- Build the Deployment object from the request payload ---
    try:
        deployment_payload = {
            "deploymentId": req.deployment.deployment_id,
            "resources": req.deployment.resources,
        }
        deployment_obj = undump_deployment(deployment_payload)
    except (KeyError, ValidationError) as exc:
        return JSONResponse(
            {"detail": f"Failed to parse deployment: {exc}"},
            status_code=422,
        )
    except Exception:
        return JSONResponse(
            {"detail": "Failed to parse deployment payload"},
            status_code=422,
        )

    # --- Set up the queue, cancellation event, and sequence counter ---
    # A bounded queue.Queue is thread-safe — the worker thread puts events
    # on it, and the async generator reads them off.  The maxsize applies
    # backpressure: if the consumer falls behind, the worker blocks on put()
    # rather than letting memory grow without limit.
    event_queue: queue.Queue[dict | None] = queue.Queue(maxsize=256)
    # The cancelled event is set by the async generator when the client
    # disconnects.  The worker checks it between resource actions so it
    # can stop early instead of finishing a deployment nobody is listening to.
    cancelled = threading.Event()
    seq = itertools.count(1)  # Thread-safe monotonic counter starting at 1

    def next_seq() -> int:
        """Return the next sequence number (monotonically increasing from 1)."""
        return next(seq)

    def enqueue_event(*, event_type: str, status: str, data: dict | None = None):
        """Build an event, redact it, and put it on the queue.

        If the client has disconnected (cancelled is set), silently drops
        the event instead of blocking on a full queue that nobody reads.
        """
        if cancelled.is_set():
            return
        event = make_event(
            deployment_id=deployment_id,
            request_id=request_id,
            correlation_id=correlation_id,
            event_type=event_type,
            status=status,
            sequence=next_seq(),
            data=data,
        )
        event_queue.put(redact_event(event, secrets))

    # --- Worker function (runs in a daemon thread) ---
    def worker():
        """Run the deployment and emit events.

        This runs in a separate thread because deploy.run() is synchronous
        and would block the async event loop if called directly.
        """
        try:
            # Create a fresh httpx client for this deployment
            client = fbnconfig.create_client(request.state.lusid_env, request.state.token)

            enqueue_event(event_type="deployment_status", status="pending")
            enqueue_event(event_type="deployment_status", status="running")

            worker_start = time.perf_counter()

            # The event_cb is called by deploy.run() for each resource action.
            # It checks the cancelled event — if the client has disconnected
            # we raise to abort the deployment at the next resource boundary.
            def event_cb(evt: dict):
                if cancelled.is_set():
                    raise _DeployCancelledError()
                enqueue_event(
                    event_type=evt["eventType"],
                    status=evt["status"],
                    data=evt["data"],
                )

            actions = deploy_run(client, deployment_obj, event_cb=event_cb)

            # Build summary from the actions list
            duration_ms = round(
                (time.perf_counter() - worker_start) * 1000, 3
            )
            action_counts: dict[str, int] = {}
            for action in actions:
                action_counts[action.change] = action_counts.get(action.change, 0) + 1

            enqueue_event(
                event_type="summary",
                status="success",
                data={"actions": action_counts, "durationMs": duration_ms},
            )
            enqueue_event(event_type="deployment_status", status="success")

        except _DeployCancelledError:
            _json_log(
                logging.INFO,
                "deploy_stream_cancelled",
                request_id=request_id,
                correlation_id=correlation_id,
                deployment_id=deployment_id,
            )

        except Exception as exc:
            error_data = _map_exception(exc)
            enqueue_event(event_type="error", status="failure", data=error_data)
            enqueue_event(event_type="deployment_status", status="failure")

        finally:
            # Signal "no more events" to the reader
            event_queue.put(_STREAM_SENTINEL)
            with _active_workers_lock:
                if thread in _active_workers:
                    _active_workers.remove(thread)

    # --- Async generator that reads from the queue ---
    async def event_generator() -> AsyncIterator[str]:
        """Yield JSONL lines until the worker signals completion.

        anyio.to_thread.run_sync(event_queue.get) is the bridge between
        the sync queue and the async world — it runs queue.get() in a
        thread pool so it doesn't block the event loop.

        The finally block sets the cancelled event so the worker thread
        knows to stop if the client disconnects mid-stream.
        """
        try:
            while True:
                event = await to_thread.run_sync(event_queue.get)
                if event is None:
                    break
                yield to_jsonl_line(event)
        finally:
            cancelled.set()
            duration_ms = round(
                (time.perf_counter() - stream_start) * 1000, 3
            )
            _json_log(
                logging.INFO,
                "deploy_stream_finished",
                request_id=request_id,
                correlation_id=correlation_id,
                deployment_id=deployment_id,
                duration_ms=duration_ms,
            )

    # --- Start the worker thread and return the streaming response ---
    thread = threading.Thread(target=worker)
    thread.start()
    with _active_workers_lock:
        _active_workers.append(thread)

    return StreamingResponse(
        event_generator(),
        media_type="application/x-ndjson",
        headers={
            "Cache-Control": "no-store",
            "X-Accel-Buffering": "no",
        },
    )


SCHEMA = schemagen.cmd_deployment_schema()


def not_found_error_handler(request, exc: NotFoundError):
    return JSONResponse(
        status_code=http.HTTPStatus.NOT_FOUND,
        content={"detail": exc.message},
    )


def validation_error_handler(request, exc: ValidationError):
    return JSONResponse(
        status_code=http.HTTPStatus.UNPROCESSABLE_ENTITY,
        content={"detail": exc.errors()},
    )


def http_status_exception_handler(request, exc: HTTPStatusError):
    return JSONResponse(
        status_code=exc.response.status_code,
        content={"detail": exc.response.text} if exc.response.status_code not in [
            http.HTTPStatus.NOT_FOUND, http.HTTPStatus.UNAUTHORIZED
        ] else None,
    )


# Wrapping every request to capture timing, add tracing IDs, and log structured data.
class RequestLogger(BaseHTTPMiddleware):

    # Constants for duration calculation
    MS_IN_SECOND = 1000
    DURATION_PRECISION = 3

    async def dispatch(self, request: Request, call_next):

        # Extract existing IDs from headers or generate new ones for traceability
        request_id, correlation_id = _get_request_ids(request)

        # Store IDs in request state so downstream handlers can access them
        request.state.request_id = request_id
        request.state.correlation_id = correlation_id

        # Record start time for duration calculation
        start_time = time.perf_counter()

        try:
            response = await call_next(request)
        except Exception as exc:
            # Log the error before re-raising so we capture failed requests
            self._log_request(
                level=logging.ERROR,
                event="request_error",
                request=request,
                request_id=request_id,
                correlation_id=correlation_id,
                error=str(exc),
                start_time=start_time
            )
            raise

        # Add correlation ID to response headers for client-side tracing
        if correlation_id:
            response.headers["Correlationid"] = correlation_id

        # Log successful request completions
        self._log_request(
            level=logging.INFO,
            event="request_complete",
            request=request,
            request_id=request_id,
            correlation_id=correlation_id,
            status=response.status_code,
            start_time=start_time
        )
        return response

    def _log_request(self, *, level, event, request, request_id, correlation_id, start_time, **extra):
        """Log request details in structured JSON for observability."""
        duration_ms = round(
            (time.perf_counter() - start_time) * self.MS_IN_SECOND, self.DURATION_PRECISION
        )
        _json_log(
            level,
            event,
            request_id=request_id,
            correlation_id=correlation_id,
            method=request.method,
            path=request.url.path,
            duration_ms=duration_ms,
            **extra
        )


class PassThruAuth(BaseHTTPMiddleware):
    async def dispatch(self, request: Request, call_next):
        auth = request.headers.get("Authorization")
        if auth and auth.startswith("Bearer "):
            request.state.token = auth.split(" ")[1]
        else:
            request.state.token = None
        host = request.headers.get("X-LUSID-Host")
        request.state.lusid_env = host if host else None
        return await call_next(request)


middleware = [Middleware(RequestLogger), Middleware(PassThruAuth)]
routes = [
    Route("/", homepage),
    Route("/health", health_check),
    Route("/ready", ready_check),
    Route("/invoke", invoke, methods=["POST"]),
    Route("/examples/", list_examples),
    Route("/examples/{example_name}", get_example),
    Route("/log/{deployment_name}", get_log),
    Route("/schema", get_schema, methods=["GET"]),

    Route("/resources/metadata", get_metadata, methods=["GET"]),
    Route("/resources/{resource}/list", get_list, methods=["POST"]),
    Route("/resources/{resource}/fetch", get_fetch, methods=["POST"]),
    Route("/deploy/stream", deploy_stream, methods=["POST"]),
]

# the fbnconfig application
cfg_app = Starlette(
    routes=routes,
    middleware=middleware,
    exception_handlers=cast(Any, {
        NotFoundError: not_found_error_handler,
        ValidationError: validation_error_handler,
        HTTPStatusError: http_status_exception_handler,
    }),
)

# the app that will be run to mount the cfg app
app = Starlette(routes=[
    Mount("/api/fbnconfig", cfg_app)
])


@app.router.on_event("startup")
async def startup():
    """Record startup time and log structured event for observability."""
    app.state.started_at = datetime.now(timezone.utc)
    try:
        _json_log(
            logging.INFO,
            "app_startup",
            timestamp=app.state.started_at.isoformat(),
            version=getattr(fbnconfig, "__version__", "unknown")
        )
    except Exception as exc:
        logger.error(f"Startup logging failed: {exc}")


@app.router.on_event("shutdown")
async def shutdown():
    """Drain active deployment workers before shutting down."""
    with _active_workers_lock:
        threads = list(_active_workers)

    if threads:
        _json_log(
            logging.INFO,
            "app_shutdown_draining",
            active_workers=len(threads),
            timeout_s=_WORKER_DRAIN_TIMEOUT_S,
        )
        for t in threads:
            t.join(timeout=_WORKER_DRAIN_TIMEOUT_S)

    try:
        _json_log(
            logging.INFO,
            "app_shutdown",
            timestamp=datetime.now(timezone.utc).isoformat(),
        )
    except Exception as exc:
        logger.error(f"Shutdown logging failed: {exc}")
